# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# Assume unicode literals as per Python 3
from __future__ import unicode_literals

import six
import hszinc
from copy import copy, deepcopy

def check_singleton_deepcopy(S):
    orig_dict = {'some_value': S}
    copy_dict = deepcopy(orig_dict)
    assert copy_dict['some_value'] is S

def test_marker_deepcopy():
    check_singleton_deepcopy(hszinc.MARKER)

def test_remove_deepcopy():
    check_singleton_deepcopy(hszinc.REMOVE)

def check_singleton_copy(S):
    assert copy(S) is S

def test_marker_copy():
    check_singleton_copy(hszinc.MARKER)

def test_remove_copy():
    check_singleton_copy(hszinc.REMOVE)

def test_ref_notref_eq():
    r1 = hszinc.Ref(name='a.ref')
    r2 = 'not.a.ref'
    assert r1 is not r2
    assert not (r1 == r2)

def test_ref_notref_ne():
    r1 = hszinc.Ref(name='a.ref')
    r2 = 'not.a.ref'
    assert r1 is not r2
    assert r1 != r2

def test_ref_simple_eq():
    r1 = hszinc.Ref(name='a.ref')
    r2 = hszinc.Ref(name='a.ref')
    assert r1 is not r2
    assert r1 == r2

def test_ref_simple_neq_id():
    r1 = hszinc.Ref(name='a.ref')
    r2 = hszinc.Ref(name='another.ref')
    assert r1 is not r2
    assert r1 != r2

def test_ref_mixed_neq():
    r1 = hszinc.Ref(name='a.ref')
    r2 = hszinc.Ref(name='a.ref', value='display text')
    assert r1 is not r2
    assert r1 != r2

def test_ref_withdis_eq():
    r1 = hszinc.Ref(name='a.ref', value='display text')
    r2 = hszinc.Ref(name='a.ref', value='display text')
    assert r1 is not r2
    assert r1 == r2

def test_ref_withdis_neq_id():
    r1 = hszinc.Ref(name='a.ref', value='display text')
    r2 = hszinc.Ref(name='another.ref', value='display text')
    assert r1 is not r2
    assert r1 != r2

def test_ref_withdis_neq_dis():
    r1 = hszinc.Ref(name='a.ref', value='display text')
    r2 = hszinc.Ref(name='a.ref', value='different display text')
    assert r1 is not r2
    assert r1 != r2

def test_qty_unary_ops():
    # Try this both without, and with, pint enabled
    for pint_en in (False, True):
        hszinc.use_pint(pint_en)
        # How to run the test: check the result
        # applied to the Quantity object matches what
        # would be returned for the same operation applied
        # to the raw value.
        def _check_qty_op(fn, *vals):
            for v in vals:
                q = hszinc.Quantity(v)
                assert fn(q) == fn(q.value)

        # These work for floats
        for fn in ( lambda v : int(v),
                    lambda v : complex(v),
                    lambda v : float(v),
                    lambda v : -v,
                    lambda v : +v,
                    lambda v : abs(v)):
            yield _check_qty_op, fn, 123.45, -123.45

        # These work for integers
        for fn in ( lambda v : oct(v),
                    lambda v : hex(v),
                    lambda v : v.__index__(),
                    lambda v : int(v),
                    lambda v : complex(v),
                    lambda v : float(v),
                    lambda v : -v,
                    lambda v : +v,
                    lambda v : abs(v),
                    lambda v : ~v):
            yield _check_qty_op, fn, 123, -123

        # This only works with Python 2.
        if six.PY2:
            yield _check_qty_op, lambda v : long(v), 123.45, -123.45

def test_qty_hash():
    # Test that independent copies hash to the same value
    def _check_hash(val, unit=None):
        a = hszinc.Quantity(val, unit=unit)
        b = hszinc.Quantity(val, unit=unit)

        assert a is not b
        assert hash(a) == hash(b)

    # Try this both without, and with, pint enabled
    for pint_en in (False, True):
        hszinc.use_pint(pint_en)

        _check_hash(123.45)
        _check_hash(-123.45)
        _check_hash(12345)
        _check_hash(-12345)
        _check_hash(50, 'Hz')

def test_qty_binary_ops():
    def _check_qty_op(fn, a, b):
        qa = hszinc.Quantity(a)
        qb = hszinc.Quantity(b)

        # Reference value
        ref = fn(a, b)

        assert fn(qa, qb) == ref
        assert fn(qa, b) == ref
        assert fn(a, qb) == ref

    for pint_en in (False, True):
        hszinc.use_pint(pint_en)
        # Try some float values
        floats = (1.12, 2.23, -4.56, 141.2, -399.5)
        for a in floats:
            for b in floats:
                if a == b:
                    continue

                for fn in ( lambda a, b : a + b,
                            lambda a, b : a - b,
                            lambda a, b : a * b,
                            lambda a, b : a / b,
                            lambda a, b : a // b,
                            lambda a, b : a % b,
                            lambda a, b : divmod(a, b),
                            lambda a, b : a < b,
                            lambda a, b : a <= b,
                            lambda a, b : a == b,
                            lambda a, b : a != b,
                            lambda a, b : a >= b,
                            lambda a, b : a > b):
                    yield _check_qty_op, fn, a, b

        # Exponentiation, we can't use all the values above
        # as some go out of range.
        small_floats = tuple(filter(lambda f : abs(f) < 10, floats))
        for a in small_floats:
            for b in small_floats:
                if a == b:
                    continue

                # Python2 won't allow raising negative numbers
                # to a fractional power
                if a < 0:
                    continue

                yield _check_qty_op, lambda a, b: a ** b, a, b

        # Try some integer values
        ints = (1, 2, -4, 141, -399, 0x10, 0xff, 0x55)
        for a in ints:
            for b in ints:
                if a == b:
                    continue

                for fn in ( lambda a, b : a + b,
                            lambda a, b : a - b,
                            lambda a, b : a * b,
                            lambda a, b : a / b,
                            lambda a, b : a // b,
                            lambda a, b : a % b,
                            lambda a, b : divmod(a, b),
                            lambda a, b : a & b,
                            lambda a, b : a ^ b,
                            lambda a, b : a | b,
                            lambda a, b : a < b,
                            lambda a, b : a <= b,
                            lambda a, b : a == b,
                            lambda a, b : a != b,
                            lambda a, b : a >= b,
                            lambda a, b : a > b):
                    yield _check_qty_op, fn, a, b

                if b >= 0:
                    for fn in ( lambda a, b : a << b,
                                lambda a, b : a >> b):
                        yield _check_qty_op, fn, a, b

        # Exponentiation, we can't use all the values above
        # as some go out of range.
        small_ints = tuple(filter(lambda f : abs(f) < 10, ints))
        for a in small_ints:
            for b in small_ints:
                if a == b:
                    continue

                yield _check_qty_op, lambda a, b: a ** b, a, b

def test_qty_cmp():
    if 'cmp' not in set(locals().keys()):
        def cmp(a, b):
            return a.__cmp__(b)

    for pint_en in (False, True):
        hszinc.use_pint(pint_en)

        a = hszinc.Quantity(-3)
        b = hszinc.Quantity(432)
        c = hszinc.Quantity(4, unit='A')
        d = hszinc.Quantity(10, unit='A')
        e = hszinc.Quantity(12, unit='V')

        assert cmp(a, b) < 0
        assert cmp(b, a) > 0
        assert cmp(a, hszinc.Quantity(-3)) == 0
        assert cmp(c, d) < 0
        assert cmp(d, c) > 0
        assert cmp(c, hszinc.Quantity(4, unit='A')) == 0

        try:
            cmp(c, e)
        except TypeError as ex:
            assert str(ex) == 'Quantity units differ: A vs V'


class MyCoordinate(object):
    """
    A dummy class that can compare itself to a Coordinate from hszinc.
    """
    def __init__(self, lat, lng):
        self.lat = lat
        self.lng = lng

    def __eq__(self, other):
        if isinstance(other, hszinc.Coordinate):
            return (self.lat == other.latitude)\
                    and (self.lng == other.longitude)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, hszinc.Coordinate):
            return (self.lat != other.latitude)\
                    and (self.lng != other.longitude)
        return NotImplemented


def test_coord_eq():
    assert hszinc.Coordinate(latitude=33.77, longitude=-77.45) \
            == hszinc.Coordinate(latitude=33.77, longitude=-77.45)

def test_coord_eq_notcoord():
    assert not (hszinc.Coordinate(latitude=33.77, longitude=-77.45) \
            == (33.77, -77.45))

def test_coord_eq_mycoord():
    hsc = hszinc.Coordinate(latitude=33.77, longitude=-77.45)
    mc = MyCoordinate(33.77, -77.45)
    assert hsc == mc
    assert mc == hsc

def test_coord_ne():
    assert hszinc.Coordinate(latitude=-33.77, longitude=77.45) \
            != hszinc.Coordinate(latitude=33.77, longitude=-77.45)

def test_coord_ne_notcoord():
    assert (hszinc.Coordinate(latitude=33.77, longitude=-77.45) \
            != (33.77, -77.45))

def test_coord_ne_mycoord():
    hsc = hszinc.Coordinate(latitude=33.77, longitude=-77.45)
    mc = MyCoordinate(-33.77, 77.45)
    assert hsc != mc
    assert mc != hsc



# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import hszinc
import datetime
import pytz
import json

from .test_parser import SIMPLE_EXAMPLE, SIMPLE_EXAMPLE_JSON, \
        METADATA_EXAMPLE_JSON

# The metadata example is a little different, as we generate the grid without
# spaces around the commas.
METADATA_EXAMPLE='''ver:"2.0" database:"test" dis:"Site Energy Summary"
siteName dis:"Sites",val dis:"Value" unit:"kW"
"Site 1",356.214kW
"Site 2",463.028kW
'''

def make_simple_grid(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    grid.column['firstName'] = {}
    grid.column['bday'] = {}
    grid.extend([
        {
            'firstName': 'Jack',
            'bday': datetime.date(1973,7,23),
        },
        {
            'firstName': 'Jill',
            'bday': datetime.date(1975,11,15),
        },
    ])
    return grid

def test_simple():
    grid = make_simple_grid()
    grid_str = hszinc.dump(grid)
    assert grid_str == SIMPLE_EXAMPLE

def test_simple_json():
    grid = make_simple_grid()
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == SIMPLE_EXAMPLE_JSON

def make_metadata_grid(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    grid.metadata['database'] = 'test'
    grid.metadata['dis'] = 'Site Energy Summary'
    grid.column['siteName'] = {'dis': 'Sites'}
    grid.column['val'] = hszinc.MetadataObject()
    grid.column['val']['dis'] = 'Value'
    grid.column['val']['unit'] = 'kW'
    grid.extend([
        {
            'siteName': 'Site 1',
            'val': hszinc.Quantity(356.214,'kW'),
        },
        {
            'siteName': 'Site 2',
            'val': hszinc.Quantity(463.028,'kW'),
        },
    ])
    return grid

def test_metadata():
    grid = make_metadata_grid()
    grid_str = hszinc.dump(grid)
    assert grid_str == METADATA_EXAMPLE

def test_metadata_json():
    grid = make_metadata_grid()
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == METADATA_EXAMPLE_JSON

def test_multi_grid():
    grids = [make_simple_grid(), make_metadata_grid()]
    grid_str = hszinc.dump(grids)

    assert grid_str == '\n'.join([SIMPLE_EXAMPLE, METADATA_EXAMPLE])

def test_multi_grid_json():
    grids = [make_simple_grid(), make_metadata_grid()]
    grid_json = json.loads(hszinc.dump(grids, mode=hszinc.MODE_JSON))

    assert grid_json[0] == SIMPLE_EXAMPLE_JSON
    assert grid_json[1] == METADATA_EXAMPLE_JSON

def make_grid_meta(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    grid.metadata['aString'] = 'aValue'
    grid.metadata['aNumber'] = 3.14159
    grid.metadata['aNull'] = None
    grid.metadata['aMarker'] = hszinc.MARKER
    grid.metadata['aQuantity'] = hszinc.Quantity(123,'Hz')
    grid.column['empty'] = {}
    return grid

def test_grid_meta():
    grid_str = hszinc.dump(make_grid_meta())
    assert grid_str == '''ver:"2.0" aString:"aValue" aNumber:3.14159 aNull:N aMarker aQuantity:123Hz
empty
'''

def test_grid_meta_json():
    grid_json = json.loads(hszinc.dump(make_grid_meta(),
        mode=hszinc.MODE_JSON))
    assert grid_json == {
            'meta': {
                'ver': '2.0',
                'aString': 's:aValue',
                'aNumber': 'n:3.141590',
                'aNull': None,
                'aMarker': 'm:',
                'aQuantity': 'n:123.000000 Hz',
            },
            'cols': [
                {'name': 'empty'},
            ],
            'rows': [],
    }

def make_col_meta(version=hszinc.VER_2_0):
    grid = hszinc.Grid(version=version)
    col_meta = hszinc.MetadataObject()
    col_meta['aString'] = 'aValue'
    col_meta['aNumber'] = 3.14159
    col_meta['aNull'] = None
    col_meta['aMarker'] = hszinc.MARKER
    col_meta['aQuantity'] = hszinc.Quantity(123,'Hz')
    grid.column['empty'] = col_meta
    return grid

def test_col_meta():
    grid_str = hszinc.dump(make_col_meta())
    assert grid_str == '''ver:"2.0"
empty aString:"aValue" aNumber:3.14159 aNull:N aMarker aQuantity:123Hz
'''

def test_col_meta_json():
    grid_json = json.loads(hszinc.dump(make_col_meta(),
        mode=hszinc.MODE_JSON))
    assert grid_json == {
            'meta': {
                'ver': '2.0',
            },
            'cols': [
                {   'name': 'empty',
                    'aString': 's:aValue',
                    'aNumber': 'n:3.141590',
                    'aNull': None,
                    'aMarker': 'm:',
                    'aQuantity': 'n:123.000000 Hz',
                },
            ],
            'rows': [],
    }

def test_data_types_v2():
    grid = hszinc.Grid(version=hszinc.VER_2_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'A null value',
            'value': None,
        },
        {
            'comment': 'A marker',
            'value': hszinc.MARKER,
        },
        {
            'comment': 'A "remove" object',
            'value': hszinc.REMOVE,
        },
        {
            'comment': 'A boolean, indicating False',
            'value': False,
        },
        {
            'comment': 'A boolean, indicating True',
            'value': True,
        },
        {
            'comment': 'A reference, without value',
            'value': hszinc.Ref('a-ref'),
        },
        {
            'comment': 'A reference, with value',
            'value': hszinc.Ref('a-ref', 'a value'),
        },
        {
            'comment': 'A binary blob',
            'value': hszinc.Bin('text/plain'),
        },
        {
            'comment': 'A quantity',
            'value': hszinc.Quantity(500,'miles'),
        },
        {
            'comment': 'A quantity without unit',
            'value': hszinc.Quantity(500,None),
        },
        {
            'comment': 'A coordinate',
            'value': hszinc.Coordinate(-27.4725,153.003),
        },
        {
            'comment': 'A URI',
            'value': hszinc.Uri(u'http://www.example.com#`unicode:\u1234\u5678`'),
        },
        {
            'comment': 'A string',
            'value':    u'This is a test\n'\
                        u'Line two of test\n'\
                        u'\tIndented with "quotes", \\backslashes\\ and '\
                        u'Unicode characters: \u1234\u5678 and a $ dollar sign',
        },
        {
            'comment': 'A date',
            'value': datetime.date(2016,1,13),
        },
        {
            'comment': 'A time',
            'value': datetime.time(7,51,43,microsecond=12345),
        },
        {
            'comment': 'A timestamp (non-UTC)',
            'value': pytz.timezone('Europe/Berlin').localize(\
                    datetime.datetime(2016,1,13,7,51,42,12345)),
        },
        {
            'comment': 'A timestamp (UTC)',
            'value': pytz.timezone('UTC').localize(\
                    datetime.datetime(2016,1,13,7,51,42,12345)),
        },
    ])
    grid_str = hszinc.dump(grid)
    ref_str = '''ver:"2.0"
comment,value
"A null value",N
"A marker",M
"A \\"remove\\" object",R
"A boolean, indicating False",F
"A boolean, indicating True",T
"A reference, without value",@a-ref
"A reference, with value",@a-ref "a value"
"A binary blob",Bin(text/plain)
"A quantity",500miles
"A quantity without unit",500
"A coordinate",C(-27.472500,153.003000)
"A URI",`http://www.example.com#\\`unicode:\\u1234\\u5678\\``
"A string","This is a test\\nLine two of test\\n\\tIndented with \\"quotes\\", \\\\backslashes\\\\ and Unicode characters: \\u1234\\u5678 and a \\$ dollar sign"
"A date",2016-01-13
"A time",07:51:43.012345
"A timestamp (non-UTC)",2016-01-13T07:51:42.012345+01:00 Berlin
"A timestamp (UTC)",2016-01-13T07:51:42.012345+00:00 UTC
'''
    assert grid_str == ref_str

def test_data_types_v3():
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'An empty list',
            'value': [],
        },
        {
            'comment': 'A null value in a list',
            'value': [None],
        },
        {
            'comment': 'A marker in a list',
            'value': [hszinc.MARKER],
        },
        {
            'comment': 'Booleans',
            'value': [True, False],
        },
        {
            'comment': 'References',
            'value': [hszinc.Ref('a-ref'), hszinc.Ref('a-ref', 'a value')],
        },
        {
            'comment': 'A quantity',
            'value': [hszinc.Quantity(500,'miles')],
        },
    ])
    grid_str = hszinc.dump(grid)
    ref_str = '''ver:"3.0"
comment,value
"An empty list",[]
"A null value in a list",[N]
"A marker in a list",[M]
"Booleans",[T,F]
"References",[@a-ref,@a-ref "a value"]
"A quantity",[500miles]
'''
    assert grid_str == ref_str

def test_data_types_json():
    grid = hszinc.Grid(version=hszinc.VER_2_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'A null value',
            'value': None,
        },
        {
            'comment': 'A marker',
            'value': hszinc.MARKER,
        },
        {
            'comment': 'A boolean, indicating False',
            'value': False,
        },
        {
            'comment': 'A boolean, indicating True',
            'value': True,
        },
        {
            'comment': 'A reference, without value',
            'value': hszinc.Ref('a-ref'),
        },
        {
            'comment': 'A reference, with value',
            'value': hszinc.Ref('a-ref', 'a value'),
        },
        {
            'comment': 'A binary blob',
            'value': hszinc.Bin('text/plain'),
        },
        {
            'comment': 'A quantity',
            'value': hszinc.Quantity(500,'miles'),
        },
        {
            'comment': 'A quantity without unit',
            'value': hszinc.Quantity(500,None),
        },
        {
            'comment': 'A coordinate',
            'value': hszinc.Coordinate(-27.4725,153.003),
        },
        {
            'comment': 'A URI',
            'value': hszinc.Uri('http://www.example.com'),
        },
        {
            'comment': 'A string',
            'value':    'This is a test\n'\
                        'Line two of test\n'\
                        '\tIndented with "quotes" and \\backslashes\\',
        },
        {
            'comment': 'A date',
            'value': datetime.date(2016,1,13),
        },
        {
            'comment': 'A time',
            'value': datetime.time(7,51,43,microsecond=12345),
        },
        {
            'comment': 'A timestamp (non-UTC)',
            'value': pytz.timezone('Europe/Berlin').localize(\
                    datetime.datetime(2016,1,13,7,51,42,12345)),
        },
        {
            'comment': 'A timestamp (UTC)',
            'value': pytz.timezone('UTC').localize(\
                    datetime.datetime(2016,1,13,7,51,42,12345)),
        },
    ])
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == {
            'meta': {'ver':'2.0'},
            'cols': [
                {'name':'comment'},
                {'name':'value'},
            ],
            'rows': [
                {   'comment': 's:A null value',
                    'value': None},
                {   'comment': 's:A marker',
                    'value': 'm:'},
                {   'comment': 's:A boolean, indicating False',
                    'value': False},
                {   'comment': 's:A boolean, indicating True',
                    'value': True},
                {   'comment': 's:A reference, without value',
                    'value': 'r:a-ref'},
                {   'comment': 's:A reference, with value',
                    'value': 'r:a-ref a value'},
                {   'comment': 's:A binary blob',
                    'value': 'b:text/plain'},
                {   'comment': 's:A quantity',
                    'value': 'n:500.000000 miles'},
                {   'comment': 's:A quantity without unit',
                    'value': 'n:500.000000'},
                {   'comment': 's:A coordinate',
                    'value': 'c:-27.472500,153.003000'},
                {   'comment': 's:A URI',
                    'value': 'u:http://www.example.com'},
                {   'comment': 's:A string',
                    'value': 's:This is a test\n'\
                            'Line two of test\n'\
                            '\tIndented with \"quotes\" '\
                            'and \\backslashes\\'},
                {   'comment': 's:A date',
                    'value': 'd:2016-01-13'},
                {   'comment': 's:A time',
                    'value': 'h:07:51:43.012345'},
                {   'comment': 's:A timestamp (non-UTC)',
                    'value': 't:2016-01-13T07:51:42.012345+01:00 Berlin'},
                {   'comment': 's:A timestamp (UTC)',
                    'value': 't:2016-01-13T07:51:42.012345+00:00 UTC'},
            ],
    }

def test_data_types_json_v3():
    grid = hszinc.Grid(version=hszinc.VER_3_0)
    grid.column['comment'] = {}
    grid.column['value'] = {}
    grid.extend([
        {
            'comment': 'An empty list',
            'value': [],
        },
        {
            'comment': 'A null value in a list',
            'value': [None],
        },
        {
            'comment': 'A marker in a list',
            'value': [hszinc.MARKER],
        },
        {
            'comment': 'Booleans',
            'value': [True, False],
        },
        {
            'comment': 'References',
            'value': [hszinc.Ref('a-ref'), hszinc.Ref('a-ref', 'a value')],
        },
        {
            'comment': 'A quantity',
            'value': [hszinc.Quantity(500,'miles')],
        },
    ])
    grid_json = json.loads(hszinc.dump(grid, mode=hszinc.MODE_JSON))
    assert grid_json == {
            'meta': {
                'ver': '3.0'
            },
            'cols': [
                {'name': 'comment'},
                {'name': 'value'},
            ],
            'rows': [
                {
                    'comment':"s:An empty list",
                    'value':[]
                },
                {
                    'comment':"s:A null value in a list",
                    'value': [None]
                },
                {
                    'comment':"s:A marker in a list",
                    'value': ['m:']
                },
                {
                    'comment':"s:Booleans",
                    'value': [True, False]
                },
                {
                    'comment':"s:References",
                    'value': ['r:a-ref' , 'r:a-ref a value']
                },
                {
                    'comment':"s:A quantity",
                    'value': ['n:500.000000 miles'] # Python is more precise
                                                    # than The Proclaimers
                }
            ]
    }

def test_list_zinc_v2():
    try:
        grid = hszinc.Grid(version=hszinc.VER_2_0)
        grid.column['comment'] = {}
        grid.column['value'] = {}
        grid.extend([
            {
                'comment': 'An empty list',
                'value': [],
            }
        ])
        hszinc.dump(grid, mode=hszinc.MODE_ZINC)
        assert False, 'Project Haystack 2.0 doesn\'t support lists'
    except ValueError:
        pass

def test_list_json_v2():
    try:
        grid = hszinc.Grid(version=hszinc.VER_2_0)
        grid.column['comment'] = {}
        grid.column['value'] = {}
        grid.extend([
            {
                'comment': 'An empty list',
                'value': [],
            }
        ])
        hszinc.dump(grid, mode=hszinc.MODE_JSON)
        assert False, 'Project Haystack 2.0 doesn\'t support lists'
    except ValueError:
        pass

def test_scalar_zinc():
    # No need to be exhaustive, the underlying function is tested heavily by
    # the grid dump tests.
    assert hszinc.dump_scalar(hszinc.Ref('areference', 'a display name'),
            mode=hszinc.MODE_ZINC) == '@areference "a display name"'

def test_scalar_zinc_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar(["a list is not allowed in v2.0"],
                mode=hszinc.MODE_ZINC, version=hszinc.VER_2_0)
        assert False, 'Serialised a list in Haystack v2.0'
    except ValueError:
        pass

def test_scalar_json():
    # No need to be exhaustive, the underlying function is tested heavily by
    # the grid dump tests.
    assert hszinc.dump_scalar(hszinc.Ref('areference', 'a display name'),
            mode=hszinc.MODE_JSON) == 'r:areference a display name'

def test_scalar_json_ver():
    # Test that versions are respected.
    try:
        hszinc.dump_scalar(["a list is not allowed in v2.0"],
                mode=hszinc.MODE_JSON, version=hszinc.VER_2_0)
        assert False, 'Serialised a list in Haystack v2.0'
    except ValueError:
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid dumper
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .grid import Grid
from .parser import MODE_ZINC, MODE_JSON
from .version import LATEST_VER

from .zincdumper import dump_grid as dump_zinc_grid, \
                        dump_scalar as dump_zinc_scalar
from .jsondumper import dump_grid as dump_json_grid, \
                        dump_scalar as dump_json_scalar
import functools

def dump(grids, mode=MODE_ZINC):
    """
    Dump the given grids in the specified over-the-wire format.
    """
    if isinstance(grids, Grid):
        return dump_grid(grids, mode=mode)
    _dump = functools.partial(dump_grid, mode=mode)
    if mode == MODE_ZINC:
        return '\n'.join(map(_dump, grids))
    elif mode == MODE_JSON:
        return '[%s]' % ','.join(map(_dump, grids))
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

def dump_grid(grid, mode=MODE_ZINC):
    if mode == MODE_ZINC:
        return dump_zinc_grid(grid)
    elif mode == MODE_JSON:
        return dump_json_grid(grid)
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)

def dump_scalar(scalar, mode=MODE_ZINC, version=LATEST_VER):
    if mode == MODE_ZINC:
        return dump_zinc_scalar(scalar, version=version)
    elif mode == MODE_JSON:
        return dump_json_scalar(scalar, version=version)
    else: # pragma: no cover
        raise NotImplementedError('Format not implemented: %s' % mode)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Project Haystack version definitions.
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

"""
Project Haystack Version comparison.

Project Haystack does not use a standard versioning scheme like Semantic
Versioning, rather, they have a Major.Minor scheme.  This tries to mimic
the Debian-style versioning handling:

The version number itself consists of digits grouped by decimal points;
followed by some arbitrary text.

Each group of digits is compared directly.
The arbitrary text is compared lexographically.

So, in this implementation:

        2.0 < 3.0
        2.0 == 2.0.0 == 2.0.0.0
        2.0 < 2.0a
        2.0a < 2.0b
"""

import re
import warnings

VERSION_RE = re.compile(r'^(\d[\d\.]*)([^\d].*)*$')

class Version(object):
    """
    A Project Haystack version number
    """

    def __init__(self, ver_str):
        if isinstance(ver_str, Version):
            # Clone constructor
            self.version_nums = ver_str.version_nums
            self.version_extra = ver_str.version_extra
        else:
            match = VERSION_RE.match(ver_str)
            if match is None:
                raise ValueError('Not a valid version string: %r' % ver_str)

            # We should have a nice friendly dotted decimal, followed
            # by anything else not recognised.  Parse out the first bit.
            (version_nums, version_extra) = match.groups()
            self.version_nums = tuple([int(p or 0) \
                    for p in version_nums.split('.')])
            self.version_extra = version_extra

    def __str__(self):
        base = '.'.join([str(p) for p in self.version_nums])
        if self.version_extra:
            base += self.version_extra
        return base

    def _cmp(self, other):
        """
        Compare two Project Haystack version strings, then return
            -1 if self < other,
            0 if self == other
            or 1 if self > other.
        """
        if not isinstance(other, Version):
            other = Version(other)

        num1 = self.version_nums
        num2 = other.version_nums

        # Pad both to be the same length
        ver_len = max(len(num1), len(num2))
        num1 += tuple([0 for n in range(len(num1), ver_len)])
        num2 += tuple([0 for n in range(len(num2), ver_len)])

        # Compare the versions
        for (p1, p2) in zip(num1, num2):
            if p1 < p2:
                return -1
            elif p1 > p2:
                return 1

        # All the same, compare the extra strings.
        # If a version misses the extra part; we consider that as coming *before*.
        if self.version_extra is None:
            if other.version_extra is None:
                return 0
            else:
                return -1
        elif other.version_extra is None:
            return 1
        elif self.version_extra == other.version_extra:
            return 0
        elif self.version_extra < other.version_extra:
            return -1
        else:
            return 1

    def __hash__(self):
        return hash(str(self))

    # Comparison operators

    def __lt__(self, other):
        return self._cmp(other) < 0
    
    def __le__(self, other):
        return self._cmp(other) < 1

    def __eq__(self, other):
        return self._cmp(other) == 0

    def __ne__(self, other):
        return self._cmp(other) != 0

    def __ge__(self, other):
        return self._cmp(other) > -1

    def __gt__(self, other):
        return self._cmp(other) > 0

    @classmethod
    def nearest(self, ver):
        """
        Retrieve the official version nearest the one given.
        """
        if not isinstance(ver, Version):
            ver = Version(ver)

        if ver in OFFICIAL_VERSIONS:
            return ver

        # We might not have an exact match for that.
        # See if we have one that's newer than the grid we're looking at.
        versions = list(OFFICIAL_VERSIONS)
        versions.sort(reverse=True)
        best = None
        for candidate in versions:
            # Due to ambiguities, we might have an exact match and not know it.
            # '2.0' will not hash to the same value as '2.0.0', but both are
            # equivalent.
            if candidate == ver:
                # We can't beat this, make a note of the match for later
                return candidate

            # If we have not seen a better candidate, and this is older
            # then we may have to settle for that.
            if (best is None) and (candidate < ver):
                warnings.warn('This version of hszinc does not yet '\
                            'support version %s, please seek a newer version '\
                            'or file a bug.  Closest (older) version supported is %s.'\
                            % (ver, candidate))
                return candidate

            # Probably the best so far, but see if we can go closer
            if candidate > ver:
                best = candidate

        # Unhappy path, no best option?  This should not happen.
        assert best is not None
        warnings.warn('This version of hszinc does not yet '\
                    'support version %s, please seek a newer version '\
                    'or file a bug.  Closest (newer) version supported is %s.'\
                    % (ver, best))
        return best


# Standard Project Haystack versions
VER_2_0 = Version('2.0')
VER_3_0 = Version('3.0')
LATEST_VER = VER_3_0

OFFICIAL_VERSIONS = set([
    VER_2_0, VER_3_0
])



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc data types
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import six

from . import PINT_AVAILABLE

if PINT_AVAILABLE:
    from . import ureg
    from .pintutil import to_pint

from abc import ABCMeta

STR_SUB  = [
    ('\b',  '\\b'),
    ('\f',  '\\f'),
    ('\n',  '\\n'),
    ('\r',  '\\r'),
    ('\t',  '\\t'),
]

# Will keep in memory the way we want Quantity being created
MODE_PINT = False
def use_pint(val = True):
    global MODE_PINT
    if val:
        #print('Switching to Pint')
        if PINT_AVAILABLE:
            MODE_PINT = True
        else: # pragma: no cover
            # Really difficult to test this case in CI
            raise ImportError(
                    'Pint not installed. Use pip install pint if needed')
    else:
        #print('Back to default Quantity')
        MODE_PINT = False


class Quantity(six.with_metaclass(ABCMeta, object)):
    def __new__(self, value, unit=None):
        if MODE_PINT:
            return PintQuantity(value, to_pint(unit))
        else:
            return BasicQuantity(value, unit)

class Qty(object):
    """
    A quantity is a scalar value (floating point) with a unit.
    """
    def __init__(self, value, unit):
        self.value = value
        self.unit = unit

    def __repr__(self):
        return '%s(%r, %r)' % (
                self.__class__.__name__, self.value, self.unit
        )

    def __str__(self):
        return '%s %s' % (
                self.value, self.unit
        )

    def __index__(self):
        return self.value.__index__()

    def __oct__(self):
        return oct(self.value)

    def __hex__(self):
        return hex(self.value)

    def __int__(self):
        return int(self.value)

    if six.PY2: # pragma: no cover
        # Python 3 doesn't have 'long'
        def __long__(self):
            return long(self.value)

    def __complex__(self):
        return complex(self.value)

    def __float__(self):
        return float(self.value)

    def __neg__(self):
        return -self.value

    def __pos__(self):
        return +self.value

    def __abs__(self):
        return abs(self.value)

    def __invert__(self):
        return ~self.value

    def __add__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value + other

    def __sub__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value - other

    def __mul__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value * other

    def __div__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value / other

    def __truediv__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value / other

    def __floordiv__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value // other

    def __mod__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value % other

    def __divmod__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return divmod(self.value, other)

    def __pow__(self, other, modulo=None):
        if isinstance(other, Qty):
            other = other.value
        return pow(self.value, other, modulo)

    def __lshift__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value << other

    def __rshift__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value >> other

    def __and__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value & other

    def __xor__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value ^ other

    def __or__(self, other):
        if isinstance(other, Qty):
            other = other.value
        return self.value | other

    def __radd__(self, other):
        if isinstance(other, Qty): # pragma: no cover
            # Unlikely due to Qty supporting these ops directly
            other = other.value
        return other + self.value

    def __rsub__(self, other):
        if isinstance(other, Qty): # pragma: no cover
            # Unlikely due to Qty supporting these ops directly
            other = other.value
        return other - self.value

    def __rmul__(self, other):
        if isinstance(other, Qty): # pragma: no cover
            # Unlikely due to Qty supporting these ops directly
            other = other.value
        return other * self.value

    def __rdiv__(self, other):
        if isinstance(other, Qty): # pragma: no cover
            # Unlikely due to Qty supporting these ops directly
            other = other.value
        return other / self.value

    def __rtruediv__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other / self.value

    def __rfloordiv__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other // self.value

    def __rmod__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other % self.value

    def __rdivmod__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return divmod(other, self.value)

    def __rpow__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return pow(other, self.value)

    def __rlshift__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other << self.value

    def __rrshift__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other >> self.value

    def __rand__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other & self.value

    def __rxor__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other ^ self.value

    def __ror__(self, other): # pragma: no cover
        # Unlikely due to Qty supporting these ops directly
        if isinstance(other, Qty):
            other = other.value
        return other | self.value

    def _cmp_op(self, other, op):
        if isinstance(other, Qty):
            if other.unit != self.unit:
                raise TypeError('Quantity units differ: %s vs %s' \
                        % (self.unit, other.unit))
            return op(self.value, other.value)
        return op(self.value, other)

    def __lt__(self, other):
        return self._cmp_op(other, lambda x, y : x < y)

    def __le__(self, other):
        return self._cmp_op(other, lambda x, y : x <= y)

    def __eq__(self, other):
        return self._cmp_op(other, lambda x, y : x == y)

    def __ge__(self, other):
        return self._cmp_op(other, lambda x, y : x >= y)

    def __gt__(self, other):
        return self._cmp_op(other, lambda x, y : x > y)

    def __ne__(self, other):
        return self._cmp_op(other, lambda x, y : x != y)

    def __cmp__(self, other):
        if self == other:
            return 0
        elif self < other:
            return -1
        else:
            return 1

    def __hash__(self):
        return hash((self.value, self.unit))

class BasicQuantity(Qty):
    """
    Default class to be used to define Quantity.
    """
    pass

Quantity.register(BasicQuantity)

if PINT_AVAILABLE:
    class PintQuantity(Qty, ureg.Quantity):
        """
        A quantity is a scalar value (floating point) with a unit.
        This object uses Pint feature allowing conversion between units
        for example :
            a = hszinc.Q_(19, 'degC')
            a.to('degF')
        See https://pint.readthedocs.io for details
        """
        def __init__(self, value, unit):
            super(PintQuantity, self).__init__(value, unit)
    Quantity.register(PintQuantity)
else: # pragma: no cover
    # If things turn really bad...just in case.
    PintQuantity = BasicQuantity
    to_pint = lambda unit: unit


class Coordinate(object):
    """
    A 2D co-ordinate in degrees latitude and longitude.
    """
    def __init__(self, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude

    def __repr__(self):
        return '%s(%r, %r)' % (
                self.__class__.__name__, self.latitude, self.longitude
        )

    def __str__(self):
        return '%f° lat %f° long' % (
                self.latitude, self.longitude
        )

    def __eq__(self, other):
        if not isinstance(other, Coordinate):
            return NotImplemented
        return (self.latitude == other.latitude) and \
                (self.longitude == other.longitude)

    def __ne__(self, other):
        if not isinstance(other, Coordinate):
            return NotImplemented
        return not (self == other)


class Uri(six.text_type):
    """
    A convenience class to allow identification of a URI from other string
    types.
    """
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                        super(Uri, self).__repr__())

    def __eq__(self, other):
        if not isinstance(other, Uri):
            return NotImplemented
        return super(Uri, self).__eq__(other)


class Bin(six.text_type):
    """
    A convenience class to allow identification of a Bin from other string
    types.
    """
    # TODO: This seems to be the MIME type, no idea where the data lives.
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                        super(Bin, self).__repr__())

    def __eq__(self, other):
        if not isinstance(other, Bin):
            return NotImplemented
        return super(Bin, self).__eq__(other)

class Singleton(object):
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

class MarkerType(Singleton):
    """
    A singleton class representing a Marker.
    """
    def __repr__(self):
        return 'MARKER'

MARKER = MarkerType()

class RemoveType(Singleton):
    """
    A singleton class representing a Remove.
    """
    def __repr__(self):
        return 'REMOVE'

REMOVE = RemoveType()


class Ref(object):
    """
    A reference to an object in Project Haystack.
    """
    # TODO: The grammar specifies that it can have a string following a space,
    # but the documentation does not specify what this string encodes.  This is
    # distinct from the reference name itself immediately following the @
    # symbol.  I'm guessing it's some kind of value.
    def __init__(self, name, value=None, has_value=False):
        self.name       = name
        self.value      = value
        self.has_value  = has_value or (value is not None)

    def __repr__(self):
        return '%s(%r, %r, %r)' % (
                self.__class__.__name__, self.name, self.value, self.has_value
        )

    def __str__(self):
        if self.has_value:
            return '@%s %r' % (
                    self.name, self.value
            )
        else:
            return '@%s' % self.name

    def __eq__(self, other):
        if not isinstance(other, Ref):
            return NotImplemented
        return (self.name == other.name) and \
                (self.has_value == other.has_value) and \
                (self.value == other.value)

    def __ne__(self, other):
        if not isinstance(other, Ref):
            return NotImplemented
        return not (self == other)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid dumper
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from __future__ import unicode_literals

from .datatypes import Quantity, Coordinate, Ref, Bin, Uri, \
        MARKER, REMOVE, STR_SUB
from .zoneinfo import timezone_name
from .version import LATEST_VER, VER_3_0

import datetime
import re
import functools
import six

URI_META = re.compile(r'([\\`\u0080-\uffff])')
STR_META = re.compile(r'([\\"\$\u0080-\uffff])')

def str_sub(match):
    c = match.group(0)
    o = ord(c)
    if o >= 0x0080:
        # Unicode
        return '\\u%04x' % o
    elif c in '\\"$':
        return '\\%s' % c

def uri_sub(match):
    c = match.group(0)
    o = ord(c)
    if o >= 0x80:
        # Unicode
        return '\\u%04x' % o
    elif c in '\\`':
        return '\\%s' % c

def dump_grid(grid):
    """
    Dump a single grid to its ZINC representation.
    """
    header = 'ver:%s' % dump_str(str(grid._version), version=grid._version)
    if bool(grid.metadata):
        header += ' ' + dump_meta(grid.metadata, version=grid._version)
    columns = dump_columns(grid.column, version=grid._version)
    rows = dump_rows(grid)
    return '\n'.join([header, columns] + rows + [''])

def dump_meta(meta, version=LATEST_VER):
    _dump = functools.partial(dump_meta_item, version=version)
    return ' '.join(map(_dump, list(meta.items())))

def dump_meta_item(item, version=LATEST_VER):
    (item_id, item_value) = item
    if item_value is MARKER:
        return dump_id(item_id, version=version)
    else:
        return '%s:%s' % (dump_id(item_id, version=version), \
                dump_scalar(item_value, version=version))

def dump_columns(cols, version=LATEST_VER):
    _dump = functools.partial(dump_column, version=version)
    _cols = list(zip(*list(cols.items())))
    return ','.join(map(_dump, *_cols))

def dump_column(col, col_meta, version=LATEST_VER):
    if bool(col_meta):
        return '%s %s' % (dump_id(col, version=version), \
                dump_meta(col_meta, version=version))
    else:
        return dump_id(col, version=version)

def dump_rows(grid):
    return list(map(functools.partial(dump_row, grid), grid))

def dump_row(grid, row):
    return ','.join([dump_scalar(row.get(c), version=grid.version) for \
            c in list(grid.column.keys())])

def dump_scalar(scalar, version=LATEST_VER):
    if scalar is None:
        return 'N'
    elif scalar is MARKER:
        return 'M'
    elif scalar is REMOVE:
        return 'R'
    elif isinstance(scalar, list):
        # Forbid version 2.0 and earlier.
        if version < VER_3_0:
            raise ValueError('Project Haystack version %s '\
                            'does not support lists'\
                            % version)
        return '[%s]' % ','.join(map(
                functools.partial(dump_scalar, version=version),
                scalar))
    elif isinstance(scalar, bool):
        return dump_bool(scalar, version=version)
    elif isinstance(scalar, Ref):
        return dump_ref(scalar, version=version)
    elif isinstance(scalar, Bin):
        return dump_bin(scalar, version=version)
    elif isinstance(scalar, Uri):
        return dump_uri(scalar, version=version)
    elif isinstance(scalar, six.string_types):
        return dump_str(scalar, version=version)
    elif isinstance(scalar, datetime.datetime):
        return dump_date_time(scalar, version=version)
    elif isinstance(scalar, datetime.time):
        return dump_time(scalar, version=version)
    elif isinstance(scalar, datetime.date):
        return dump_date(scalar, version=version)
    elif isinstance(scalar, Coordinate):
        return dump_coord(scalar, version=version)
    elif isinstance(scalar, Quantity):
        return dump_quantity(scalar, version=version)
    elif isinstance(scalar, float) or \
            isinstance(scalar, int) or \
            isinstance(scalar, int):
        return dump_decimal(scalar, version=version)
    else: # pragma: no cover
        raise NotImplementedError('Unhandled case: %r' % scalar)

def dump_id(id_str, version=LATEST_VER):
    return id_str

def dump_str(str_value, version=LATEST_VER):
    # Replace special characters.
    str_value = STR_META.sub(str_sub, str_value)
    # Replace other escapes.
    for orig, esc in STR_SUB:
        str_value = str_value.replace(orig, esc)
    return '"%s"' % str_value

def dump_uri(uri_value, version=LATEST_VER):
    # Replace special characters.
    uri_value = URI_META.sub(uri_sub, uri_value)
    # Replace other escapes.
    for orig, esc in STR_SUB:
        uri_value = uri_value.replace(orig, esc)
    return '`%s`' % uri_value

def dump_bin(bin_value, version=LATEST_VER):
    return 'Bin(%s)' % bin_value

def dump_quantity(quantity, version=LATEST_VER):
    if (quantity.unit is None) or (quantity.unit == ''):
        return dump_decimal(quantity.value, version=version)
    else:
        return '%s%s' % (dump_decimal(quantity.value),
                quantity.unit)

def dump_decimal(decimal, version=LATEST_VER):
    return str(decimal)

def dump_bool(bool_value, version=LATEST_VER):
    return 'T' if bool(bool_value) else 'F'

def dump_coord(coordinate, version=LATEST_VER):
    return 'C(%f,%f)' % (coordinate.latitude, coordinate.longitude)

def dump_ref(ref, version=LATEST_VER):
    if ref.has_value:
        return '@%s %s' % (ref.name, dump_str(ref.value))
    else:
        return '@%s' % ref.name

def dump_date(date, version=LATEST_VER):
    return date.isoformat()

def dump_time(time, version=LATEST_VER):
    return time.isoformat()

def dump_date_time(date_time, version=LATEST_VER):
    tz_name = timezone_name(date_time, version=version)
    return '%s %s' % (date_time.isoformat(), tz_name)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc grammar specification.
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si: 

import pyparsing as pp
import iso8601
import datetime
import string
import warnings
import six
import logging

import re

# Bring in special Project Haystack types and time zones
from .datatypes import Quantity, Coordinate, Uri, Bin, MARKER, REMOVE, Ref
from .grid import Grid
from .zoneinfo import timezone

# Bring in our sortable dict class to preserve order
from .sortabledict import SortableDict

# Bring in version handling
from .version import Version, LATEST_VER, VER_2_0, VER_3_0

# Logging instance for reporting debug info
LOG = logging.getLogger(__name__)

# All grids start with the version string.
VERSION_RE = re.compile(r'^ver:"(([^"\\]|\\[\\"bfnrt$])+)"')
NEWLINE_RE = re.compile(r'\r?\n')

# Character number regex; for exceptions
CHAR_NUM_RE = re.compile(' *\(at char \d+\),')

def reformat_exception(ex_msg, line_num=None):
    print (ex_msg)
    msg = CHAR_NUM_RE.sub(u'', six.text_type(ex_msg))
    print (msg)
    if line_num is not None:
        return msg.replace(u'line:1', u'line:%d' % line_num)
    else:
        return msg

# Convenience function, we want whitespace left alone.
def _leave_ws(cls, *args, **kwargs):
    return cls(*args, **kwargs).leaveWhitespace()

# Versions of the pyparsing types that leave our whitespace alone!
Empty           = lambda *a, **kwa : _leave_ws(pp.Empty, *a, **kwa)
Regex           = lambda *a, **kwa : _leave_ws(pp.Regex, *a, **kwa)
Literal         = lambda *a, **kwa : _leave_ws(pp.Literal, *a, **kwa)
CaselessLiteral = lambda *a, **kwa : _leave_ws(pp.CaselessLiteral, *a, **kwa)
Word            = lambda *a, **kwa : _leave_ws(pp.Word, *a, **kwa)
Optional        = lambda *a, **kwa : _leave_ws(pp.Optional, *a, **kwa)
Suppress        = lambda *a, **kwa : _leave_ws(pp.Suppress, *a, **kwa)
Combine         = lambda *a, **kwa : _leave_ws(pp.Combine, *a, **kwa)
And             = lambda *a, **kwa : _leave_ws(pp.And, *a, **kwa)
Or              = lambda *a, **kwa : _leave_ws(pp.Or, *a, **kwa)
ZeroOrMore      = lambda *a, **kwa : _leave_ws(pp.ZeroOrMore, *a, **kwa)
OneOrMore       = lambda *a, **kwa : _leave_ws(pp.OneOrMore, *a, **kwa)
Group           = lambda *a, **kwa : _leave_ws(pp.Group, *a, **kwa)
DelimitedList   = lambda *a, **kwa : _leave_ws(pp.delimitedList, *a, **kwa)
Forward         = lambda *a, **kwa : _leave_ws(pp.Forward, *a, **kwa)


class ZincParseException(ValueError):
    """
    Exception thrown when a grid cannot be parsed successfully.  If known,
    the line and column for the grid are given.
    """
    def __init__(self, message, grid_str, line, col):
        self.grid_str = grid_str
        self.line = line
        self.col = col

        try:
            # If we know the line and column, point it out in the message.
            grid_str_lines = grid_str.split('\n')
            width = max([len(l) for l in grid_str_lines])
            linefmt = u'%%-%ds' % width
            rowfmt = u'%4d%s' + linefmt + u'%s'

            formatted_lines = [
                    rowfmt % (
                        num,
                        ' >' if (line == num) else '| ',
                        line_str,
                        '< ' if (line == num) else ' |'
                    )
                    for (num, line_str)
                    in enumerate(grid_str.split('\n'), 1)
            ]
            formatted_lines.insert(line,
                    (u'    | ' + linefmt + u' |') \
                            % (((col-2) * u' ') + '.^.')
            )

            # Border it for readability
            formatted_lines.insert(0, u'    .' + (u'-' * (2 + width)) + u'.')
            formatted_lines.append(u'    \'' + (u'-' * (2 + width)) + u'\'')

            # Append to message
            message += u'\n%s' % u'\n'.join(formatted_lines)
        except: # pragma: no cover
            # We should not get here.
            LOG.exception('Exception encountered formatting log message')
            pass

        super(ZincParseException, self).__init__(message)


class NearestMatch(object):
    """
    This class returns the nearest matching grammar for the given version.
    """
    def __init__(self, known_grammars):
        self._known_grammars = known_grammars

    def __getitem__(self, ver):
        """
        Retrieve the grammar that closest matches the version string given.
        """
        try:
            return self._known_grammars[ver]
        except KeyError:
            pass

        nearest = Version.nearest(ver)
        g = self._known_grammars[nearest]
        self._known_grammars[ver] = g
        return g


class GenerateMatch(object):
    """
    This class tries to generate a matching grammar based on the version input given.
    """
    def __init__(self, generator_fn):
        self._generator_fn = generator_fn
        self._known_grammars = {}

    def __getitem__(self, ver):
        try:
            return self._known_grammars[ver]
        except KeyError:
            g = self._generator_fn(ver)
            self._known_grammars[ver] = g
            return g

def _unescape(s, uri=False):
    """
    Iterative parser for string escapes.
    """
    out = ''
    while len(s) > 0:
        c = s[0]
        if c == '\\':
            # Backslash escape
            esc_c = s[1]

            if esc_c in ('u', 'U'):
                # Unicode escape
                out += six.unichr(int(s[2:6], base=16))
                s = s[6:]
                continue
            else:
                if esc_c == 'b':
                    out += '\b'
                elif esc_c == 'f':
                    out += '\f'
                elif esc_c == 'n':
                    out += '\n'
                elif esc_c == 'r':
                    out += '\r'
                elif esc_c == 't':
                    out += '\t'
                else:
                    if uri and (esc_c == '#'):
                        # \# is passed through with backslash.
                        out += '\\'
                    # Pass through
                    out += esc_c
                s = s[2:]
                continue
        else:
            out += c
            s = s[1:]
    return out

# Grammar according to
#   latest: http://project-haystack.org/doc/Zinc
#   "2.0":  https://web.archive.org/web/20141012013653/http://project-haystack.org:80/doc/Zinc
#   "3.0":  https://web.archive.org/web/20160805064015/http://project-haystack.org:80/doc/Zinc

# Rudimentary elements
hs_digit        = Regex(r'\d')
hs_digits       = Regex(r'[0-9_]+').setParseAction(
        lambda toks : [''.join([t.replace('_','') for t in toks[0]])])
hs_alphaLo      = Regex(r'[a-z]')
hs_alphaHi      = Regex(r'[A-Z]')
hs_alpha        = Regex(r'[a-zA-Z]')
hs_valueSep     = Regex(r' *, *').setName('valueSep')
hs_plusMinus    = Or([Literal('+'), Literal('-')])

# Forward declaration of data types.
hs_scalar_2_0   = Forward()
hs_scalar_3_0   = Forward()
hs_scalar = NearestMatch({
        VER_2_0: hs_scalar_2_0,
        VER_3_0: hs_scalar_3_0
})

# Co-ordinates
hs_coordDeg     = Combine(And([
        Optional(Literal('-')),
        Optional(hs_digits),
        Optional(And([Literal('.'), hs_digits]))
])).setParseAction(lambda toks : [float(toks[0] or '0')])
hs_coord        = And([Suppress(Literal('C(')),
                hs_coordDeg,
                Suppress(hs_valueSep),
                hs_coordDeg,
                Suppress(Literal(')'))]).setParseAction(
                        lambda toks : [Coordinate(toks[0], toks[1])])

# Dates and times
hs_tzHHMMOffset = Combine(Or([
        CaselessLiteral('z'),
        And([hs_plusMinus, Regex(r'\d\d:\d\d')])]
))
hs_tzName       = Regex(r'[A-Z][a-zA-Z0-9_\-]*')
hs_tzUTCGMT     = Or([Literal('UTC'), Literal('GMT')])
hs_tzUTCOffset  = Combine(And([
    hs_tzUTCGMT, Optional(
        Or([Literal('0'),
            And([hs_plusMinus, OneOrMore(hs_digit)]
        )]
    ))]))
hs_timeZoneName = Or([hs_tzUTCOffset, hs_tzName])
hs_dateSep      = CaselessLiteral('T')
hs_date_str     = Combine(And([
    hs_digit, hs_digit, hs_digit, hs_digit,
    Literal('-'),
    hs_digit, hs_digit,
    Literal('-'),
    hs_digit, hs_digit]))
hs_date         = hs_date_str.copy().setParseAction(
        lambda toks : [datetime.datetime.strptime(toks[0], '%Y-%m-%d').date()])

hs_time_str     = Combine(And([
    hs_digit, hs_digit,
    Literal(':'),
    hs_digit, hs_digit,
    Literal(':'),
    hs_digit, hs_digit,
    Optional(And([
            Literal('.'),
            OneOrMore(hs_digit)]))
    ]))

def _parse_time(toks):
    time_str = toks[0]
    time_fmt = '%H:%M:%S'
    if '.' in time_str:
        time_fmt += '.%f'
    return [datetime.datetime.strptime(time_str, time_fmt).time()]
hs_time         = hs_time_str.copy().setParseAction(_parse_time)
hs_isoDateTime  = Combine(And([
    hs_date_str,
    hs_dateSep,
    hs_time_str,
    Optional(hs_tzHHMMOffset)
])).setParseAction(lambda toks : [iso8601.parse_date(toks[0].upper())])

def _parse_datetime(toks):
    # Made up of parts: ISO8601 Date/Time, time zone label
    isodt = toks[0]
    if len(toks) > 1:
        tzname = toks[1]
    else:
        tzname = None

    if (isodt.tzinfo is None) and bool(tzname): # pragma: no cover
        # This technically shouldn't happen according to Zinc specs
        return [timezone(tzname).localise(isodt)]
    elif bool(tzname):
        try:
            tz = timezone(tzname)
            return [isodt.astimezone(tz)]
        except: # pragma: no cover
            # Unlikely to occur, might do though if Project Haystack changes
            # its timezone list or if a system doesn't recognise a particular
            # timezone.
            return [isodt]  # Failed, leave alone
    else:
        return [isodt]
hs_dateTime     = And([
    hs_isoDateTime,
    Optional(And([
        Suppress(Literal(' ')),
        hs_timeZoneName
        ]))
    ]).setParseAction(_parse_datetime)

# Quantities and raw numeric values
hs_unitChar     = Or([
    hs_alpha,
    Word(u'%_/$' + u''.join([
        six.unichr(c)
        for c in range(0x0080, 0xffff)
    ]), exact=1)
])
hs_unit         = Combine(OneOrMore(hs_unitChar))
hs_exp          = Combine(And([
    CaselessLiteral('e'),
    Optional(hs_plusMinus),
    hs_digits
]))
hs_decimal      = Combine(And([
    Optional(Literal('-')),
    hs_digits,
    Optional(And([
        Literal('.'),
        hs_digits
    ])),
    Optional(hs_exp)
])).setParseAction(lambda toks : [float(toks[0])])

hs_quantity     = And([hs_decimal, hs_unit]).setParseAction(
        lambda toks : [Quantity(toks[0], unit=toks[1])])
hs_number       = Or([
    hs_quantity,
    hs_decimal,
    Or([
        Literal('INF'),
        Literal('-INF'),
        Literal('NaN')
    ]).setParseAction(lambda toks : [float(toks[0])])
])

# URIs
hs_uriChar      = Regex(r"([^\x00-\x1f\\`]|\\[bfnrt\\:/?" \
                        + r"#\[\]@&=;`]|\\[uU][0-9a-fA-F]{4})")
hs_uri          = Combine(And([
    Suppress(Literal('`')),
    ZeroOrMore(hs_uriChar),
    Suppress(Literal('`'))
])).setParseAction(lambda toks : [Uri(_unescape(toks[0], uri=True))])

# Strings
hs_strChar      = Regex(r"([^\x00-\x1f\\\"]|\\[bfnrt\\\"$]|\\[uU][0-9a-fA-F]{4})")
hs_str          = Combine(And([
    Suppress(Literal('"')),
    ZeroOrMore(hs_strChar),
    Suppress(Literal('"'))
])).setParseAction(lambda toks : [_unescape(toks[0], uri=False)])

# References
hs_refChar      = Or([hs_alpha, hs_digit, Word('_:-.~', exact=1)])
hs_ref          = And([
    Suppress(Literal('@')),
    Combine(ZeroOrMore(hs_refChar)),
    Optional(And([
        Suppress(Literal(' ')),
        hs_str
    ]))
]).setParseAction(lambda toks : [\
        Ref(toks[0], toks[1] if len(toks) > 1 else None)\
])

# Bins
hs_binChar      = Regex(r"[\x20-\x27\x2a-\x7f]")
hs_bin          = Combine(And([
    Suppress(Literal('Bin(')),
    Combine(ZeroOrMore(hs_binChar)),
    Suppress(Literal(')'))
])).setParseAction(lambda toks : [Bin(toks[0])])

# Booleans
hs_bool         = Word('TF', min=1, max=1, exact=1).setParseAction(\
        lambda toks : [toks[0] == 'T'])

# Singleton values
hs_remove       = Literal('R').setParseAction(\
        lambda toks : [REMOVE]).setName('remove')
hs_marker       = Literal('M').setParseAction(\
        lambda toks : [MARKER]).setName('marker')
hs_null         = Literal('N').setParseAction(\
        lambda toks : [None]).setName('null')

# Lists, these will probably be in Haystack 4.0, so let's not
# assume a version.  There are three cases:
# - Empty list: [ {optional whitespace} ]
# - List *with* trailing comma: [ 1, 2, 3, ]
# - List without trailing comma: [ 1, 2, 3 ]
#
# We need to handle this trailing separator case.  That for now means
# that a NULL within a list *MUST* be explicitly given using the 'N'
# literal: we cannot support implicit NULLs as they are ambiguous.
hs_list         = GenerateMatch(                            \
        lambda ver : Group(Or([                             \
            Suppress(Regex(r'[ *]')),                       \
            And([                                           \
                Suppress(Regex(r'\[ *')),                   \
                Optional(DelimitedList(                     \
                        hs_scalar[ver],                     \
                        delim=hs_valueSep)),                \
                Suppress(Optional(hs_valueSep)),            \
                Suppress(Regex(r' *\]'))                    \
            ])                                              \
        ])).setParseAction(lambda toks : toks.asList()))

# All possible scalar values, by Haystack version
hs_scalar_2_0 <<= Or([hs_ref, hs_bin, hs_str, hs_uri, hs_dateTime,
            hs_date, hs_time, hs_coord, hs_number, hs_null, hs_marker,
            hs_remove, hs_bool]).setName('scalar')
hs_scalar_3_0 <<= Or([hs_ref, hs_bin, hs_str, hs_uri, hs_dateTime,
            hs_date, hs_time, hs_coord, hs_number, hs_null, hs_marker,
            hs_remove, hs_bool, hs_list[VER_3_0]]).setName('scalar')

# Tag IDs
hs_id           = Regex(r'[a-z][a-zA-Z0-9_]*').setName('id')

# Grid building blocks
hs_cell         = GenerateMatch(                                                \
        lambda ver : Or([Empty().copy().setParseAction(lambda toks : [None]),   \
                        hs_scalar[ver]]).setName('cell'))
hs_nl           = Combine(And([Optional(Literal('\r')), Literal('\n')]))

hs_row          = GenerateMatch(\
        lambda ver : Group(DelimitedList(hs_cell[ver], delim=hs_valueSep)))
hs_metaPair     = GenerateMatch(\
        lambda ver : And([ \
            hs_id, \
            Suppress(And([ \
                ZeroOrMore(Literal(' ')), \
                Literal(':'), \
                ZeroOrMore(Literal(' ')) \
            ])), \
            hs_scalar[ver] \
        ]).setParseAction(lambda toks : [tuple(toks[:2])]).setName('metaPair'))
hs_metaMarker   = hs_id.copy().setParseAction(\
        lambda toks : [(toks[0], MARKER)]).setName('metaMarker')
hs_metaItem     = GenerateMatch(\
        lambda ver : Or([\
                hs_metaMarker, \
                hs_metaPair[ver]\
            ]).setName('metaItem'))
hs_meta         = GenerateMatch(\
        lambda ver : DelimitedList(hs_metaItem[ver], \
                delim=' ').setParseAction(\
                    lambda toks : [SortableDict(toks.asList())] \
            ).setName('meta'))

hs_col          = GenerateMatch(\
        lambda ver : And([ \
            hs_id, \
            Optional(And([ \
                Suppress(Literal(' ')), \
                hs_meta[ver]
            ])).setName('colMeta') \
        ]).setParseAction(lambda toks : [\
            (toks[0], toks[1] if len(toks) > 1 else {})]))

hs_cols         = GenerateMatch(\
        lambda ver : DelimitedList(
            hs_col[ver], delim=hs_valueSep).setParseAction( # + hs_nl
                lambda toks : [SortableDict(toks.asList())]))

hs_gridVer      = Combine(And([Suppress(Literal('ver:')) + hs_str]))

def _assign_ver(toks):
    ver = toks[0]
    if len(toks) > 1:
        grid_meta = toks[1]
    else:
        grid_meta = SortableDict()

    # Put 'ver' at the start
    grid_meta.add_item('ver', ver, index=0)
    return grid_meta
hs_gridMeta     = GenerateMatch(\
        lambda ver : And([ \
            hs_gridVer, \
            Optional(And([ \
                Suppress(Literal(' ')), \
                hs_meta[ver] \
            ])).setName('gridMeta')
        ]).setParseAction(_assign_ver)) # + hs_nl


def parse_grid(grid_data):
    """
    Parse the incoming grid.
    """
    # Split the grid up.
    grid_parts = NEWLINE_RE.split(grid_data)
    if len(grid_parts) < 2:
        raise ZincParseException('Malformed grid received',
                grid_data, 1, 1)

    # Grid and column metadata are the first two lines.
    grid_meta_str = grid_parts.pop(0)
    col_meta_str = grid_parts.pop(0)

    # First element is the grid metadata
    ver_match = VERSION_RE.match(grid_meta_str)
    if ver_match is None:
        raise ZincParseException(
                'Could not determine version from %r' % grid_meta_str,
                grid_data, 1, 1)
    version = Version(ver_match.group(1))

    # Now parse the rest of the grid accordingly
    try:
        grid_meta = hs_gridMeta[version].parseString(grid_meta_str, parseAll=True)[0]
    except pp.ParseException as pe:
        # Raise a new exception with the appropriate line number.
        raise ZincParseException(
                'Failed to parse grid metadata: %s' % pe,
                grid_data, 1, pe.col)
    except: # pragma: no cover
        # Report an error to the log if we fail to parse something.
        LOG.debug('Failed to parse grid meta: %r', grid_meta_str)
        raise

    try:
        col_meta = hs_cols[version].parseString(col_meta_str, parseAll=True)[0]
    except pp.ParseException as pe:
        # Raise a new exception with the appropriate line number.
        raise ZincParseException(
                'Failed to parse column metadata: %s' \
                        % reformat_exception(pe, 2),
                grid_data, 2, pe.col)
    except: # pragma: no cover
        # Report an error to the log if we fail to parse something.
        LOG.debug('Failed to parse column meta: %r', col_meta_str)
        raise

    row_grammar = hs_row[version]
    def _parse_row(row_num_and_data):
        (row_num, row) = row_num_and_data
        line_num = row_num + 3

        try:
            return dict(zip(col_meta.keys(),
                row_grammar.parseString(row, parseAll=True)[0].asList()))
        except pp.ParseException as pe:
            # Raise a new exception with the appropriate line number.
            raise ZincParseException(
                    'Failed to parse row: %s' \
                        % reformat_exception(pe, line_num),
                    grid_data, line_num, pe.col)
        except: # pragma: no cover
            # Report an error to the log if we fail to parse something.
            LOG.debug('Failed to parse row: %r', row)
            raise

    g = Grid(version=grid_meta.pop('ver'),
            metadata=grid_meta,
            columns=list(col_meta.items()))
    g.extend(map(_parse_row, filter(lambda gp : bool(gp[1]), enumerate(grid_parts))))
    return g


def parse_scalar(scalar_data, version):
    """
    Parse a Project Haystack scalar in ZINC format.
    """
    try:
        return hs_scalar[version].parseString(scalar_data, parseAll=True)[0]
    except pp.ParseException as pe:
        # Raise a new exception with the appropriate line number.
        raise ZincParseException(
                'Failed to parse scalar: %s' % reformat_exception(pe),
                scalar_data, 1, pe.col)



import hszinc
import pkg_resources, os

units = []

hszinc.use_pint()

def get_units():

    file_path = os.path.join('', 'project_haystack_units.txt')
    units_file = pkg_resources.resource_string(__name__, file_path)
    global units
    with open(units_file, 'r', encoding = 'UTF-8') as file:
        for line in file:
            if '--' not in line and line != '':
                res = line.rstrip().split(',')
                if res[0] != '':
                    units.append(res[0])
                if len(res) > 1:
                    units.append(res[1])
    assert len(units) > 0

def test_all_units():
    not_defined = []
    defined = []
    for each in units:
        try:   
            q = hszinc.Q_(1,each)
            defined.append(each)
            print(each, q)
        except Exception as error:
            not_defined.append(each)
            print(error, each)
    assert (len(not_defined)) == 0


# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

import hszinc
import datetime
import math
import pytz
import json

# These are examples taken from http://project-haystack.org/doc/Zinc

hszinc.use_pint()

SIMPLE_EXAMPLE='''ver:"2.0"
firstName,bday
"Jack",1973-07-23
"Jill",1975-11-15
'''

SIMPLE_EXAMPLE_JSON={
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'firstName'},
            {'name':'bday'},
        ],
        'rows': [
            {'firstName': 's:Jack', 'bday':'d:1973-07-23'},
            {'firstName': 's:Jill', 'bday':'d:1975-11-15'},
        ],
}

def test_simple():
    grid_list = hszinc.parse(SIMPLE_EXAMPLE)
    assert len(grid_list) == 1
    grid = grid_list[0]
    check_simple(grid)

def test_simple_json_str():
    grid = hszinc.parse(json.dumps(SIMPLE_EXAMPLE_JSON),
            mode=hszinc.MODE_JSON)
    check_simple(grid)

def check_row_keys(row, grid):
    assert set(row.keys()) == set(grid.column.keys())

def check_simple(grid):
    assert len(grid.metadata) == 0
    assert list(grid.column.keys()) == ['firstName', 'bday']
    # Neither column should have metadata
    assert all([len(c) == 0 for c in grid.column.values()])

    # First row:
    row = grid[0]
    check_row_keys(row, grid)
    assert row['firstName'] == 'Jack'
    assert row['bday'] == datetime.date(1973,7,23)

    # Second row:
    row = grid[1]
    check_row_keys(row, grid)
    assert row['firstName'] == 'Jill'
    assert row['bday'] == datetime.date(1975,11,15)

METADATA_EXAMPLE='''ver:"2.0" database:"test" dis:"Site Energy Summary"
siteName dis:"Sites", val dis:"Value" unit:"kW"
"Site 1", 356.214kW
"Site 2", 463.028kW
'''

METADATA_EXAMPLE_JSON={
        'meta': {'ver':'2.0', 'database':'s:test',
                'dis':'s:Site Energy Summary'},
        'cols': [
            {'name':'siteName', 'dis':'s:Sites'},
            {'name':'val', 'dis':'s:Value', 'unit':'s:kW'},
        ],
        'rows': [
            {'siteName': 's:Site 1', 'val': 'n:356.214000 kW'},
            {'siteName': 's:Site 2', 'val': 'n:463.028000 kW'},
        ],
}

def test_metadata():
    grid_list = hszinc.parse(METADATA_EXAMPLE)
    assert len(grid_list) == 1
    grid = grid_list[0]
    check_metadata(grid)

def test_metadata_json():
    grid = hszinc.parse(METADATA_EXAMPLE_JSON, mode=hszinc.MODE_JSON)
    check_metadata(grid, force_metadata_order=False)

def check_metadata(grid,force_metadata_order=True):
    assert len(grid.metadata) == 2
    if force_metadata_order:
        assert list(grid.metadata.keys()) == ['database', 'dis']
    else:
        assert set(grid.metadata.keys()) == set(['database', 'dis'])

    assert grid.metadata['database'] == 'test'
    assert grid.metadata['dis'] == 'Site Energy Summary'

    assert list(grid.column.keys()) == ['siteName', 'val']
    col = grid.column['siteName']
    assert list(col.keys()) == ['dis']
    assert col['dis'] == 'Sites'

    col = grid.column['val']
    if force_metadata_order:
        assert list(col.keys()) == ['dis', 'unit']
    else:
        assert set(col.keys()) == set(['dis', 'unit'])
    assert col['dis'] == 'Value'
    assert col['unit'] == 'kW'

    # First row:
    row = grid[0]
    check_row_keys(row, grid)
    assert row['siteName'] == 'Site 1'
    val = row['val']
    assert isinstance(val, hszinc.Quantity)
    assert val.value == 356.214
    assert val.unit == 'kW'

    # Second row:
    row = grid[1]
    check_row_keys(row, grid)
    assert row['siteName'] == 'Site 2'
    val = row['val']
    assert isinstance(val, hszinc.Quantity)
    assert val.value == 463.028
    assert val.unit == 'kW'

# Test examples used to test every data type defined in the Zinc standard.
#    Null: N
#    Marker: M
#    Bool: T or F (for true, false)
#    Number: 1, -34, 10_000, 5.4e-45, 9.23kg, 74.2°F, 4min, INF, -INF, NaN
#    Str: "hello" "foo\nbar\" (uses all standard escape chars as C like languages)
#    Uri: `http://project-haystack.com/`
#    Ref: @17eb0f3a-ad607713
#    Date: 2010-03-13 (YYYY-MM-DD)
#    Time: 08:12:05 (hh:mm:ss.FFF)
#    DateTime: 2010-03-11T23:55:00-05:00 New_York or 2009-11-09T15:39:00Z
#    Bin: Bin(text/plain)
#    Coord: C(37.55,-77.45)

NULL_EXAMPLE='''ver:"2.0"
str,null
"Implicit",
"Explict",N
'''
NULL_EXAMPLE_JSON={
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'str'},
            {'name':'null'},
        ],
        'rows': [
            {'str': 'Implicit', 'null':None},   # There's no "implicit" mode
            {'str': 'Explicit', 'null':None},
        ],
}


def check_null(grid):
    assert len(grid) == 2
    assert grid[0]['null'] is None
    assert grid[1]['null'] is None

def test_null():
    grid_list = hszinc.parse(NULL_EXAMPLE)
    assert len(grid_list) == 1
    check_null(grid_list[0])

def test_null_json():
    grid = hszinc.parse(NULL_EXAMPLE_JSON, mode=hszinc.MODE_JSON)
    check_null(grid)

def test_marker_in_row():
    grid_list = hszinc.parse('''ver:"2.0"
str,marker
"No Marker",
"Marker",M
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 2
    assert grid_list[0][0]['marker'] is None
    assert grid_list[0][1]['marker'] is hszinc.MARKER

def test_marker_in_row_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'str'},
            {'name':'marker'},
        ],
        'rows': [
            {'str': 'No Marker',    'marker':None},
            {'str': 'Marker',       'marker':'m:'},
        ],
    }, mode=hszinc.MODE_JSON)
    assert grid[0]['marker'] is None
    assert grid[1]['marker'] is hszinc.MARKER

def test_bool():
    grid_list = hszinc.parse('''ver:"2.0"
str,bool
"True",T
"False",F
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 2
    assert grid_list[0][0]['bool'] == True
    assert grid_list[0][1]['bool'] == False

def test_bool_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'str'},
            {'name':'bool'},
        ],
        'rows': [
            {'str': 'True', 'bool':True},
            {'str': 'False','bool':False},
        ],
    }, mode=hszinc.MODE_JSON)
    assert grid[0]['bool'] == True
    assert grid[1]['bool'] == False

def test_number():
    grid_list = hszinc.parse(u'''ver:"2.0"
str,number
"Integer",1
"Negative Integer",-34
"With Separators",10_000
"Scientific",5.4e-45
"Units mass",9.23kg
"Units time",4min
"Units temperature",74.2°F
"Positive Infinity",INF
"Negative Infinity",-INF
"Not a Number",NaN
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    check_number(grid)

def check_number(grid):
    assert len(grid) == 10
    row = grid.pop(0)
    assert row['number'] == 1.0
    row = grid.pop(0)
    assert row['number'] == -34.0
    row = grid.pop(0)
    assert row['number'] == 10000.0
    row = grid.pop(0)
    assert row['number'] == 5.4e-45
    row = grid.pop(0)
    assert row['number'] == hszinc.Quantity(9.23, 'kg')
    row = grid.pop(0)
    assert row['number'] == hszinc.Quantity(4.0, 'min')
    row = grid.pop(0)
    assert row['number'] == hszinc.Quantity(74.2, u'°F')
    row = grid.pop(0)
    assert math.isinf(row['number'])
    assert row['number'] > 0
    row = grid.pop(0)
    assert math.isinf(row['number'])
    assert row['number'] < 0
    row = grid.pop(0)
    assert math.isnan(row['number'])

def test_number_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'str'},
            {'name':'number'},
        ],
        'rows': [
            {'str': "Integer",              'number': u'n:1'},
            {'str': "Negative Integer",     'number': u'n:-34'},
            {'str': "With Separators",      'number': u'n:10000'},
            {'str': "Scientific",           'number': u'n:5.4e-45'},
            {'str': "Units mass",           'number': u'n:9.23 kg'},
            {'str': "Units time",           'number': u'n:4 min'},
            {'str': "Units temperature",    'number': u'n:74.2 °F'},
            {'str': "Positive Infinity",    'number': u'n:INF'},
            {'str': "Negative Infinity",    'number': u'n:-INF'},
            {'str': "Not a Number",         'number': u'n:NaN'},
        ],
    }, mode=hszinc.MODE_JSON)
    check_number(grid)

def test_string():
    grid_list = hszinc.parse('''ver:"2.0"
str,strExample
"Empty",""
"Basic","Simple string"
"Escaped","This\\tIs\\nA\\r\\"Test\\"\\\\"
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 3
    assert grid_list[0][0]['strExample'] == ''
    assert grid_list[0][1]['strExample'] == 'Simple string'
    assert grid_list[0][2]['strExample'] == 'This\tIs\nA\r"Test"\\'

def test_string_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'str'},
            {'name':'strExample'},
        ],
        'rows': [
            {'str': "Empty",            'strExample': ''},
            {'str': "Implicit",         'strExample': 'a string'},
            {'str': "Literal",          'strExample': 's:an explicit string'},
            {'str': "With colons",      'strExample': 's:string:with:colons'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 4
    assert grid.pop(0)['strExample'] == ''
    assert grid.pop(0)['strExample'] == 'a string'
    assert grid.pop(0)['strExample'] == 'an explicit string'
    assert grid.pop(0)['strExample'] == 'string:with:colons'

def test_uri():
    grid_list = hszinc.parse('''ver:"2.0"
uri
`http://www.vrt.com.au`
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert grid_list[0][0]['uri'] == hszinc.Uri('http://www.vrt.com.au')

def test_uri_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'uri'},
        ],
        'rows': [
            {'uri': 'u:http://www.vrt.com.au'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 1
    assert grid[0]['uri'] == hszinc.Uri('http://www.vrt.com.au')

def test_ref():
    grid_list = hszinc.parse('''ver:"2.0"
str,ref
"Basic",@a-basic-ref
"With value",@reference "With value"
''')

    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 2
    assert grid[0]['ref'] == hszinc.Ref('a-basic-ref')
    assert grid[1]['ref'] == hszinc.Ref('reference', 'With value')

def test_ref_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'str'},
            {'name':'ref'},
        ],
        'rows': [
            {'str': 'Basic',        'ref': 'r:a-basic-ref'},
            {'str': 'With value',   'ref': 'r:reference With value'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 2
    assert grid[0]['ref'] == hszinc.Ref('a-basic-ref')
    assert grid[1]['ref'] == hszinc.Ref('reference', 'With value')

def test_date():
    grid_list = hszinc.parse('''ver:"2.0"
date
2010-03-13
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert isinstance(grid_list[0][0]['date'], datetime.date)
    assert grid_list[0][0]['date'] == datetime.date(2010,3,13)

def test_date_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'date'},
        ],
        'rows': [
            {'date': 'd:2010-03-13'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 1
    assert isinstance(grid[0]['date'], datetime.date)
    assert grid[0]['date'] == datetime.date(2010,3,13)

def test_time():
    grid_list = hszinc.parse('''ver:"2.0"
time
08:12:05
08:12:05.5
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 2
    assert isinstance(grid_list[0][0]['time'], datetime.time)
    assert grid_list[0][0]['time'] == datetime.time(8,12,5)
    assert isinstance(grid_list[0][1]['time'], datetime.time)
    assert grid_list[0][1]['time'] == datetime.time(8,12,5,500000)

def test_time_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'time'},
        ],
        'rows': [
            {'time': 'h:08:12'},
            {'time': 'h:08:12:05'},
            {'time': 'h:08:12:05.5'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 3
    row = grid.pop(0)
    assert isinstance(row['time'], datetime.time)
    assert row['time'] == datetime.time(8,12)
    row = grid.pop(0)
    assert isinstance(row['time'], datetime.time)
    assert row['time'] == datetime.time(8,12,5)
    row = grid.pop(0)
    assert isinstance(row['time'], datetime.time)
    assert row['time'] == datetime.time(8,12,5,500000)

def test_datetime():
    grid_list = hszinc.parse('''ver:"2.0"
datetime
2010-11-28T07:23:02.500-08:00 Los_Angeles
2010-11-28T23:19:29.500+08:00 Taipei
2010-11-28T18:21:58+03:00 GMT-3
2010-11-28T12:22:27-03:00 GMT+3
2010-01-08T05:00:00Z UTC
2010-01-08T05:00:00Z
''')

    assert len(grid_list) == 1
    check_datetime(grid_list.pop(0))

def check_datetime(grid):
    assert len(grid) == 6
    row = grid.pop(0)
    assert isinstance(row['datetime'], datetime.datetime)
    assert row['datetime'] == \
            pytz.timezone('America/Los_Angeles').localize(\
            datetime.datetime(2010,11,28,7,23,2,500000))
    row = grid.pop(0)
    assert row['datetime'] == \
            pytz.timezone('Asia/Taipei').localize(\
            datetime.datetime(2010,11,28,23,19,29,500000))
    row = grid.pop(0)
    assert row['datetime'] == \
            pytz.timezone('Etc/GMT-3').localize(\
            datetime.datetime(2010,11,28,18,21,58,0))
    row = grid.pop(0)
    assert row['datetime'] == \
            pytz.timezone('Etc/GMT+3').localize(\
            datetime.datetime(2010,11,28,12,22,27,0))
    row = grid.pop(0)
    assert row['datetime'] == \
            pytz.timezone('UTC').localize(\
            datetime.datetime(2010,1,8,5,0,0,0))
    row = grid.pop(0)
    assert row['datetime'] == \
            pytz.timezone('UTC').localize(\
            datetime.datetime(2010,1,8,5,0,0,0))

def test_datetime_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'datetime'},
        ],
        'rows': [
            {'datetime': 't:2010-11-28T07:23:02.500-08:00 Los_Angeles'},
            {'datetime': 't:2010-11-28T23:19:29.500+08:00 Taipei'},
            {'datetime': 't:2010-11-28T18:21:58+03:00 GMT-3'},
            {'datetime': 't:2010-11-28T12:22:27-03:00 GMT+3'},
            {'datetime': 't:2010-01-08T05:00:00Z UTC'},
            {'datetime': 't:2010-01-08T05:00:00Z'},
        ],
    }, mode=hszinc.MODE_JSON)
    check_datetime(grid)

def test_bin():
    grid_list = hszinc.parse('''ver:"2.0"
bin
Bin(text/plain)
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert grid_list[0][0]['bin'] == hszinc.Bin('text/plain')

def test_bin_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'bin'},
        ],
        'rows': [
            {'bin': 'b:text/plain'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 1
    assert grid[0]['bin'] == hszinc.Bin('text/plain')

def test_coord():
    grid_list = hszinc.parse('''ver:"2.0"
coord
C(37.55,-77.45)
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert grid_list[0][0]['coord'] == hszinc.Coordinate(37.55,-77.45)

def test_coord_json():
    grid = hszinc.parse({
        'meta': {'ver':'2.0'},
        'cols': [
            {'name':'coord'},
        ],
        'rows': [
            {'coord': 'c:37.55,-77.45'},
        ],
    }, mode=hszinc.MODE_JSON)

    assert len(grid) == 1
    assert grid[0]['coord'] == hszinc.Coordinate(37.55,-77.45)

def test_multi_grid():
    # Multiple grids are separated by newlines.
    grid_list = hszinc.parse('\n'.join([
        SIMPLE_EXAMPLE, METADATA_EXAMPLE, NULL_EXAMPLE]))
    assert len(grid_list) == 3
    check_simple(grid_list[0])
    check_metadata(grid_list[1])
    check_null(grid_list[2])

def test_multi_grid_json():
    # Multiple grids are separated by newlines.
    grid_list = hszinc.parse([SIMPLE_EXAMPLE_JSON,
        METADATA_EXAMPLE_JSON, NULL_EXAMPLE_JSON], mode=hszinc.MODE_JSON)
    assert len(grid_list) == 3
    check_simple(grid_list[0])
    check_metadata(grid_list[1], force_metadata_order=False)
    check_null(grid_list[2])

def test_multi_grid_json_str():
    # Multiple grids are separated by newlines.
    grid_list = hszinc.parse(list(map(json.dumps, [SIMPLE_EXAMPLE_JSON,
        METADATA_EXAMPLE_JSON, NULL_EXAMPLE_JSON])), mode=hszinc.MODE_JSON)
    assert len(grid_list) == 3
    check_simple(grid_list[0])
    check_metadata(grid_list[1], force_metadata_order=False)
    check_null(grid_list[2])

def test_grid_meta():
    grid_list = hszinc.parse('''ver:"2.0" aString:"aValue" aNumber:3.14159 aNull:N aMarker:M anotherMarker aQuantity:123Hz aDate:2016-01-13 aTime:06:44:00 aTimestamp:2016-01-13T06:44:00+10:00 Brisbane aPlace:C(-27.4725,153.003)
empty
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 0
    meta = grid_list[0].metadata
    assert list(meta.keys()) == ['aString', 'aNumber', 'aNull',
            'aMarker', 'anotherMarker', 'aQuantity', 'aDate', 'aTime',
            'aTimestamp', 'aPlace']
    assert meta['aString'] == 'aValue'
    assert meta['aNumber'] == 3.14159
    assert meta['aNull'] is None
    assert meta['aMarker'] is hszinc.MARKER
    assert meta['anotherMarker'] is hszinc.MARKER
    assert isinstance(meta['aQuantity'], hszinc.Quantity)
    assert meta['aQuantity'].value == 123
    assert meta['aQuantity'].unit == 'Hz'
    assert isinstance(meta['aDate'], datetime.date)
    assert meta['aDate'] == datetime.date(2016,1,13)
    assert isinstance(meta['aTime'], datetime.time)
    assert meta['aTime'] == datetime.time(6,44)
    assert isinstance(meta['aTimestamp'], datetime.datetime)
    assert meta['aTimestamp'] == \
            pytz.timezone('Australia/Brisbane').localize(\
                datetime.datetime(2016,1,13,6,44))
    assert isinstance(meta['aPlace'], hszinc.Coordinate)
    assert meta['aPlace'].latitude == -27.4725
    assert meta['aPlace'].longitude == 153.003

def test_col_meta():
    grid_list = hszinc.parse('''ver:"2.0"
aColumn aString:"aValue" aNumber:3.14159 aNull:N aMarker:M anotherMarker aQuantity:123Hz aDate:2016-01-13 aTime:06:44:00 aTimestamp:2016-01-13T06:44:00+10:00 Brisbane aPlace:C(-27.4725,153.003)
''')

    assert len(grid_list) == 1
    assert len(grid_list[0]) == 0
    assert len(grid_list[0].metadata) == 0
    assert list(grid_list[0].column.keys()) == ['aColumn']
    meta = grid_list[0].column['aColumn']
    assert list(meta.keys()) == ['aString', 'aNumber', 'aNull',
            'aMarker', 'anotherMarker', 'aQuantity', 'aDate', 'aTime',
            'aTimestamp', 'aPlace']
    assert meta['aString'] == 'aValue'
    assert meta['aNumber'] == 3.14159
    assert meta['aNull'] is None
    assert meta['aMarker'] is hszinc.MARKER
    assert meta['anotherMarker'] is hszinc.MARKER
    assert isinstance(meta['aQuantity'], hszinc.Quantity)
    assert meta['aQuantity'].value == 123
    assert meta['aQuantity'].unit == 'Hz'
    assert isinstance(meta['aDate'], datetime.date)
    assert meta['aDate'] == datetime.date(2016,1,13)
    assert isinstance(meta['aTime'], datetime.time)
    assert meta['aTime'] == datetime.time(6,44)
    assert isinstance(meta['aTimestamp'], datetime.datetime)
    assert meta['aTimestamp'] == \
            pytz.timezone('Australia/Brisbane').localize(\
                datetime.datetime(2016,1,13,6,44))
    assert isinstance(meta['aPlace'], hszinc.Coordinate)
    assert meta['aPlace'].latitude == -27.4725
    assert meta['aPlace'].longitude == 153.003

def test_too_many_cells():
    grid_list = hszinc.parse('''ver:"2.0"
col1, col2, col3
"Val1", "Val2", "Val3", "Val4", "Val5"
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert len(grid_list[0].metadata) == 0
    assert list(grid_list[0].column.keys()) == ['col1','col2','col3']

    row = grid_list[0][0]
    assert set(row.keys()) == set(['col1','col2','col3'])
    assert row['col1'] == 'Val1'
    assert row['col2'] == 'Val2'
    assert row['col3'] == 'Val3'

def test_nodehaystack_01():
    grid_list = hszinc.parse('''ver:"2.0"
fooBar33
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 0
    assert len(grid_list[0].metadata) == 0
    assert list(grid_list[0].column.keys()) == ['fooBar33']

def test_nodehaystack_02():
    grid_list = hszinc.parse('''ver:"2.0" tag foo:"bar"
xyz
"val"
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert list(grid_list[0].metadata.keys()) == ['tag', 'foo']
    assert grid_list[0].metadata['tag'] is hszinc.MARKER
    assert grid_list[0].metadata['foo'] == 'bar'
    assert list(grid_list[0].column.keys()) == ['xyz']
    assert grid_list[0][0]['xyz'] == 'val'

def test_nodehaystack_03():
    grid_list = hszinc.parse('''ver:"2.0"
val
N
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 1
    assert len(grid_list[0].metadata) == 0
    assert list(grid_list[0].column.keys()) == ['val']
    assert grid_list[0][0]['val'] is None

def test_nodehaystack_04():
    grid_list = hszinc.parse('''ver:"2.0"
a,b
1,2
3,4
''')
    assert len(grid_list) == 1
    assert len(grid_list[0]) == 2
    assert len(grid_list[0].metadata) == 0
    assert list(grid_list[0].column.keys()) == ['a','b']
    assert grid_list[0][0]['a'] == 1
    assert grid_list[0][0]['b'] == 2
    assert grid_list[0][1]['a'] == 3
    assert grid_list[0][1]['b'] == 4

def test_nodehaystack_05():
    grid_list = hszinc.parse('''ver:"2.0"
a,    b,      c,      d
T,    F,      N,   -99
2.3,  -5e-10, 2.4e20, 123e-10
"",   "a",   "\\" \\\\ \\t \\n \\r", "\\uabcd"
`path`, @12cbb082-0c02ae73, 4s, -2.5min
M,R,Bin(image/png),Bin(image/png)
2009-12-31, 23:59:01, 01:02:03.123, 2009-02-03T04:05:06Z
INF, -INF, "", NaN
C(12,-34),C(0.123,-.789),C(84.5,-77.45),C(-90,180)
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 8
    assert len(grid.metadata) == 0
    assert list(grid.column.keys()) == ['a','b','c','d']
    row = grid.pop(0)
    assert row['a'] == True
    assert row['b'] == False
    assert row['c'] is None
    assert row['d'] == -99.0
    row = grid.pop(0)
    assert row['a'] == 2.3
    assert row['b'] == -5e-10
    assert row['c'] == 2.4e20
    assert row['d'] == 123e-10
    row = grid.pop(0)
    assert row['a'] == ''
    assert row['b'] == 'a'
    assert row['c'] == '\" \\ \t \n \r'
    assert row['d'] == u'\uabcd'
    row = grid.pop(0)
    assert row['a'] == hszinc.Uri('path')
    assert row['b'] == hszinc.Ref('12cbb082-0c02ae73')
    assert row['c'] == hszinc.Quantity(4, 's')
    assert row['d'] == hszinc.Quantity(-2.5, 'min')
    row = grid.pop(0)
    assert row['a'] is hszinc.MARKER
    assert row['b'] is hszinc.REMOVE
    assert row['c'] == hszinc.Bin('image/png')
    assert row['d'] == hszinc.Bin('image/png')
    row = grid.pop(0)
    assert row['a'] == datetime.date(2009,12,31)
    assert row['b'] == datetime.time(23,59,1)
    assert row['c'] == datetime.time(1,2,3,123000)
    assert row['d'] == \
            datetime.datetime(2009,2,3,4,5,6,tzinfo=pytz.utc)
    row = grid.pop(0)
    assert math.isinf(row['a']) and (row['a'] > 0)
    assert math.isinf(row['b']) and (row['b'] < 0)
    assert row['c'] == ''
    assert math.isnan(row['d'])
    row = grid.pop(0)
    assert row['a'] == hszinc.Coordinate(12,-34)
    assert row['b'] == hszinc.Coordinate(.123,-.789)
    assert row['c'] == hszinc.Coordinate(84.5,-77.45)
    assert row['d'] == hszinc.Coordinate(-90,180)

def test_nodehaystack_06():
    grid_list = hszinc.parse('''ver:"2.0"
foo
`foo$20bar`
`foo\\`bar`
`file \\#2`
"$15"
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 4
    assert len(grid.metadata) == 0
    assert list(grid.column.keys()) == ['foo']
    row = grid.pop(0)
    assert row['foo'] == hszinc.Uri('foo$20bar')
    row = grid.pop(0)
    assert row['foo'] == hszinc.Uri('foo`bar')
    row = grid.pop(0)
    assert row['foo'] == hszinc.Uri('file \\#2')
    row = grid.pop(0)
    assert row['foo'] == '$15'


def test_nodehaystack_07():
    grid_list = hszinc.parse(u'''ver:"2.0"
a, b
-3.1kg,4kg
5%,3.2%
5kWh/ft\u00b2,-15kWh/m\u00b2
123e+12kJ/kg_dry,74\u0394\u00b0F
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 4
    assert len(grid.metadata) == 0
    assert list(grid.column.keys()) == ['a','b']
    row = grid.pop(0)
    assert row['a'] == hszinc.Quantity(-3.1,'kg')
    assert row['b'] == hszinc.Quantity(4,'kg')
    row = grid.pop(0)
    assert row['a'] == hszinc.Quantity(5,'%')
    assert row['b'] == hszinc.Quantity(3.2,'%')
    row = grid.pop(0)
    assert row['a'] == hszinc.Quantity(5,u'kWh/ft\u00b2')
    assert row['b'] == hszinc.Quantity(-15,u'kWh/m\u00b2')
    row = grid.pop(0)
    assert row['a'] == hszinc.Quantity(123e12,'kJ/kg_dry')
    assert row['b'] == hszinc.Quantity(74,u'\u0394\u00b0F')

def test_nodehaystack_08():
    grid_list = hszinc.parse('''ver:"2.0"
a,b
2010-03-01T23:55:00.013-05:00 GMT+5,2010-03-01T23:55:00.013+10:00 GMT-10
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 1
    assert len(grid.metadata) == 0
    assert list(grid.column.keys()) == ['a','b']
    row = grid.pop(0)
    assert row['a'] == hszinc.zoneinfo.timezone('GMT+5').localize(\
            datetime.datetime(2010,3,1,23,55,0,13000))
    assert row['b'] == hszinc.zoneinfo.timezone('GMT-10').localize(\
            datetime.datetime(2010,3,1,23,55,0,13000))

def test_nodehaystack_09():
    grid_list = hszinc.parse(u'''ver:"2.0" a: 2009-02-03T04:05:06Z foo b: 2010-02-03T04:05:06Z UTC bar c: 2009-12-03T04:05:06Z London baz
a
3.814697265625E-6
2010-12-18T14:11:30.924Z
2010-12-18T14:11:30.925Z UTC
2010-12-18T14:11:30.925Z London
45$
33\u00a3
@12cbb08e-0c02ae73
7.15625E-4kWh/ft\u00b2
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 8
    assert list(grid.metadata.keys()) == ['a','foo','b','bar','c','baz']
    assert grid.metadata['a'] == pytz.utc.localize(\
            datetime.datetime(2009,2,3,4,5,6))
    assert grid.metadata['b'] == pytz.utc.localize(\
            datetime.datetime(2010,2,3,4,5,6))
    assert grid.metadata['c'] == pytz.timezone('Europe/London').localize(\
            datetime.datetime(2009,12,3,4,5,6))
    assert grid.metadata['foo'] is hszinc.MARKER
    assert grid.metadata['bar'] is hszinc.MARKER
    assert grid.metadata['baz'] is hszinc.MARKER
    assert list(grid.column.keys()) == ['a']
    assert grid.pop(0)['a'] == 3.814697265625E-6
    assert grid.pop(0)['a'] == pytz.utc.localize(\
            datetime.datetime(2010,12,18,14,11,30,924000))
    assert grid.pop(0)['a'] == pytz.utc.localize(\
            datetime.datetime(2010,12,18,14,11,30,925000))
    assert grid.pop(0)['a'] == pytz.timezone('Europe/London').localize(\
            datetime.datetime(2010,12,18,14,11,30,925000))
    assert grid.pop(0)['a'] == hszinc.Quantity(45,'$')
    assert grid.pop(0)['a'] == hszinc.Quantity(33,u'\u00a3')
    assert grid.pop(0)['a'] == hszinc.Ref('12cbb08e-0c02ae73')
    assert grid.pop(0)['a'] == hszinc.Quantity(7.15625E-4,u'kWh/ft\u00b2')

def test_nodehaystack_10():
    grid_list = hszinc.parse('''ver:"2.0" bg: Bin(image/jpeg) mark
file1 dis:"F1" icon: Bin(image/gif),file2 icon: Bin(image/jpg)
Bin(text/plain),N
4,Bin(image/png)
Bin(text/html; a=foo; bar="sep"),Bin(text/html; charset=utf8)
''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 3
    assert list(grid.metadata.keys()) == ['bg','mark']
    assert grid.metadata['bg'] == hszinc.Bin('image/jpeg')
    assert grid.metadata['mark'] is hszinc.MARKER
    assert list(grid.column.keys()) == ['file1','file2']
    assert list(grid.column['file1'].keys()) == ['dis','icon']
    assert grid.column['file1']['dis'] == 'F1'
    assert grid.column['file1']['icon'] == hszinc.Bin('image/gif')
    assert list(grid.column['file2'].keys()) == ['icon']
    assert grid.column['file2']['icon'] == hszinc.Bin('image/jpg')
    row = grid.pop(0)
    assert row['file1'] == hszinc.Bin('text/plain')
    assert row['file2'] is None
    row = grid.pop(0)
    assert row['file1'] == 4
    assert row['file2'] == hszinc.Bin('image/png')
    row = grid.pop(0)
    assert row['file1'] == hszinc.Bin('text/html; a=foo; bar="sep"')
    assert row['file2'] == hszinc.Bin('text/html; charset=utf8')

def test_nodehaystack_11():
    grid_list = hszinc.parse('''ver:"2.0"
a, b, c
, 1, 2
3, , 5
6, 7_000,
,,10
,,
14,,

''')
    assert len(grid_list) == 1
    grid = grid_list.pop(0)
    assert len(grid) == 6
    assert len(grid.metadata) == 0
    assert list(grid.column.keys()) == ['a','b','c']
    row = grid.pop(0)
    assert row['a'] is None
    assert row['b'] == 1
    assert row['c'] == 2
    row = grid.pop(0)
    assert row['a'] == 3
    assert row['b'] is None
    assert row['c'] == 5
    row = grid.pop(0)
    assert row['a'] == 6
    assert row['b'] == 7000
    assert row['c'] is None
    row = grid.pop(0)
    assert row['a'] is None
    assert row['b'] is None
    assert row['c'] == 10
    row = grid.pop(0)
    assert row['a'] is None
    assert row['b'] is None
    assert row['c'] is None
    row = grid.pop(0)
    assert row['a'] == 14
    assert row['b'] is None
    assert row['c'] is None


# -*- coding: utf-8 -*-
# Zinc dumping and parsing module
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from hszinc.sortabledict import SortableDict

def test_set_get():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d[3] = 'c'
    assert d[1] == 'a'
    assert d[2] == 'b'
    assert d[3] == 'c'

def test_iter():
    d = SortableDict()
    d[3] = 'a'
    d[2] = 'b'
    d[1] = 'c'
    iter_items = list(d)
    assert iter_items == [3, 2, 1]

def test_len():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d[3] = 'c'
    assert len(d) == 3

def test_del():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d[3] = 'c'
    del d[2]
    assert d[1] == 'a'
    assert d[3] == 'c'
    assert len(d) == 2

def test_initial_dict_constructor():
    d = {
            1: 'abcd',
            "a": 1234,
    }
    assert dict(SortableDict(d)) == d

def test_initial_items_constructor():
    il = [
           (1, 'abcd'),
           ("a", 1234),
    ]
    assert list(SortableDict(il).items()) == il

def test_repr():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d[3] = 'c'
    assert repr(d) == "SortableDict{1='a', 2='b', 3='c'}"

def test_additem_with_index_and_pos_key():
    try:
        SortableDict(dict(foobar=123)).add_item('a key', 'a value', index=0, pos_key='foobar')
        assert False, 'No error raised for both key and index given'
    except ValueError as e:
        pass

def test_additem_with_invalid_pos_key():
    try:
        SortableDict().add_item('a key', 'a value', pos_key='doesnotexist')
        assert False, 'No error raised for invalid key'
    except KeyError:
        pass

def test_additem_after():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d.add_item(3, 'c', after=True, pos_key=1)
    assert list(d.items()) == [(1, 'a'), (3, 'c'), (2, 'b')]

def test_additem_noreplace_duplicate():
    d = SortableDict()
    d[1] = 'a'
    try:
        d.add_item(1, 'a2', replace=False)
        assert False, 'Replaced/inserted duplicate key'
    except KeyError:
        pass

def test_additem_move():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d[3] = 'c'
    d.add_item(1, 'a', after=True, pos_key=3)
    assert list(d.items()) == [(2,'b'), (3,'c'), (1,'a')]

def test_additem_replace():
    d = SortableDict()
    d[1] = 'a'
    d[2] = 'b'
    d[3] = 'c'
    d.add_item(1, 'd', replace=True)
    assert list(d.items()) == [(1, 'd'), (2,'b'), (3,'c')]

def test_at():
    d = SortableDict()
    d['a'] = 1
    d['b'] = 2
    d['c'] = 3
    assert d.at(0) == 'a'
    assert d.at(1) == 'b'
    assert d.at(2) == 'c'

def test_index():
    d = SortableDict()
    d['a'] = 1
    d['b'] = 2
    d['c'] = 3
    assert d.index('a') == 0
    assert d.index('b') == 1
    assert d.index('c') == 2

def test_reverse():
    d = SortableDict()
    d['a'] = 1
    d['b'] = 2
    d['c'] = 3
    d.reverse()
    assert list(d.items()) == [('c',3), ('b',2), ('a',1)]

def test_sort():
    d = SortableDict()
    d['t'] = 1
    d['d'] = 2
    d['h'] = 3
    d['x'] = 4
    d.sort()
    assert list(d.items()) == [('d',2), ('h',3), ('t',1), ('x',4)]

def test_value_at():
    d = SortableDict()
    d['a'] = 1
    d['b'] = 2
    d['c'] = 3
    assert d.value_at(0) == 1
    assert d.value_at(1) == 2
    assert d.value_at(2) == 3

def test_pop_at():
    d = SortableDict()
    d['a'] = 1
    d['b'] = 2
    d['c'] = 3
    assert d.pop_at(1) == 2
    assert list(d.items()) == [('a',1), ('c',3)]



#!/usr/bin/python
# -*- coding: utf-8 -*-
# JSON Grid Parser
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .grid import Grid
from .datatypes import Quantity, Coordinate, Ref, Bin, Uri, \
        MARKER, REMOVE, STR_SUB
from .zoneinfo import timezone
from .version import LATEST_VER, Version, VER_3_0

import datetime
import iso8601
import re
import six
import functools
import json
import copy

URI_META = re.compile(r'\\([:/\?#\[\]@\\&=;"$`])')
GRID_SEP = re.compile(r'\n\n+')

# Type regular expressions
MARKER_STR  = 'm:'
NUMBER_RE   = re.compile(r'^n:(-?\d+(:?\.\d+)?(:?[eE][+\-]?\d+)?)(:? (.*))?$',
        flags=re.MULTILINE)
REF_RE      = re.compile(r'^r:([a-zA-Z0-9_:\-.~]+)(:? (.*))?$',
        flags=re.MULTILINE)
STR_RE      = re.compile(r'^s:(.*)$', flags=re.MULTILINE)
DATE_RE     = re.compile(r'^d:(\d{4})-(\d{2})-(\d{2})$', flags=re.MULTILINE)
TIME_RE     = re.compile(r'^h:(\d{2}):(\d{2})(:?:(\d{2}(:?\.\d+)?))?$',
        flags=re.MULTILINE)
DATETIME_RE = re.compile(r'^t:(\d{4}-\d{2}-\d{2}T'\
        r'\d{2}:\d{2}(:?:\d{2}(:?\.\d+)?)'\
        r'(:?[zZ]|[+\-]\d+:?\d*))(:? ([A-Za-z\-+_0-9]+))?$',
        flags=re.MULTILINE)
URI_RE      = re.compile(r'u:(.+)$', flags=re.MULTILINE)
BIN_RE      = re.compile(r'b:(.+)$', flags=re.MULTILINE)
COORD_RE    = re.compile(r'c:(-?\d*\.?\d*),(-?\d*\.?\d*)$',
        flags=re.MULTILINE)

STR_ESC_RE  = re.compile(r'\\([bfnrt"\\$]|u[0-9a-fA-F]{4})')

def parse_grid(grid_str):
    # Grab the metadata
    if isinstance(grid_str, six.string_types):
        parsed = json.loads(grid_str)
    else:
        parsed = copy.deepcopy(grid_str)
    meta = parsed.pop('meta')
    # Decode version
    version = Version(meta.pop('ver'))

    # Parse the remaining elements
    metadata = {}
    for name, value in meta.items():
        metadata[name] = parse_scalar(value, version=version)

    grid = Grid(version=version, metadata=metadata)

    # Grab the columns in the order given
    for col in parsed.pop('cols'):
        name = col.pop('name')
        meta = {}
        for key, value in col.items():
            meta[key] = parse_scalar(value, version=version)
        grid.column[name] = meta

    # Parse the rows
    for row in (parsed.pop('rows',[]) or []):
        parsed_row = {}
        for col, value in row.items():
            parsed_row[col] = parse_scalar(value, version=version)
        grid.append(parsed_row)

    return grid


def parse_scalar(scalar, version=LATEST_VER):
    # If we're given a string, decode the JSON data.
    if isinstance(scalar, six.text_type) and \
            (len(scalar) >= 2) and \
            (scalar[0] in ('"','[','{')) and \
            (scalar[-1] in ('"',']','}')):
        scalar = json.loads(scalar)

    # Simple cases
    if scalar is None:
        return None
    elif isinstance(scalar, list):
        # We support this only in version 3.0 and up.
        if version < VER_3_0:
            raise ValueError('Lists are not supported in Haystack version %s'\
                    % version)
        return list(map(functools.partial(parse_scalar, version=version),
            scalar))
    elif scalar == MARKER_STR:
        return MARKER
    elif isinstance(scalar, bool):
        return scalar
    elif scalar == 'n:INF':
        return float('INF')
    elif scalar == 'n:-INF':
        return -float('INF')
    elif scalar == 'n:NaN':
        return float('nan')
    # Conversion to dict of float value turn them into float 
    # so regex won't work... better just return them
    elif isinstance(scalar, float) or isinstance(scalar, six.integer_types):
        return scalar

    # Is it a number?
    match = NUMBER_RE.match(scalar)
    if match:
        # We'll get a value and a unit, amongst other tokens.
        matched = match.groups()
        value = float(matched[0])
        if matched[-1] is not None:
            # It's a quantity
            return Quantity(value, matched[-1])
        else:
            # It's a raw value
            return value

    # Is it a string?
    match = STR_RE.match(scalar)
    if match:
        return match.group(1)

    # Is it a reference?
    match = REF_RE.match(scalar)
    if match:
        matched = match.groups()
        if matched[-1] is not None:
            return Ref(matched[0], matched[-1], has_value=True)
        else:
            return Ref(matched[0])

    # Is it a date?
    match = DATE_RE.match(scalar)
    if match:
        (year, month, day) = match.groups()
        return datetime.date(year=int(year), month=int(month), day=int(day))

    # Is it a time?
    match = TIME_RE.match(scalar)
    if match:
        (hour, minute, _, second, _) = match.groups()
        # Convert second to seconds and microseconds
        if second is None:
            sec = 0
            usec = 0
        elif '.' in second:
            (whole_sec, frac_sec) = second.split('.',1)
            sec = int(whole_sec)
            usec = int(frac_sec[:6].ljust(6,'0'))
        else:
            sec = int(second)
            usec = 0
        return datetime.time(hour=int(hour), minute=int(minute),
                second=sec, microsecond=usec)

    # Is it a date/time?
    match = DATETIME_RE.match(scalar)
    if match:
        matches = match.groups()
        # Parse ISO8601 component
        isodate = iso8601.parse_date(matches[0])
        # Parse timezone
        tzname  = matches[-1]
        if tzname is None:
            return isodate  # No timezone given
        else:
            try:
                tz = timezone(tzname)
                return isodate.astimezone(tz)
            except: # pragma: no cover
                # Unlikely code path.
                return isodate

    # Is it a URI?
    match = URI_RE.match(scalar)
    if match:
        return Uri(match.group(1))

    # Is it a Bin?
    match = BIN_RE.match(scalar)
    if match:
        return Bin(match.group(1))

    # Is it a co-ordinate?
    match = COORD_RE.match(scalar)
    if match:
        (lat,lng) = match.groups()
        return Coordinate(float(lat),float(lng))
    return scalar



# -*- coding: utf-8 -*-
# Grid handling tests
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# Assume unicode literals as per Python 3
from __future__ import unicode_literals

from hszinc import Grid, MARKER, Version
from hszinc.sortabledict import SortableDict

def test_grid_given_metadata():
    # Test that when passing in metadata, it is set as given.
    meta = SortableDict()
    meta['first'] = 1
    meta['second'] = 2
    meta['third'] = 3
    meta['fourth'] = 4

    assert list(Grid(metadata=meta).metadata.items()) == [
            ('first', 1), ('second', 2), ('third', 3), ('fourth', 4)
    ]

def test_grid_given_column_list():
    col_list = [('col1', [('c1m1', None), ('c1m2', None)]),
                ('col2', [('c2m1', None), ('c2m2', None)])]
    g = Grid(columns=col_list)
    assert list(g.column.keys()) == ['col1','col2']
    for col, meta in col_list:
        assert list(g.column[col].items()) == meta

def test_grid_given_column_dict():
    cols = SortableDict([('col1', [('c1m1', None), ('c1m2', None)]),
                        ('col2', [('c2m1', None), ('c2m2', None)])])
    g = Grid(columns=cols)
    assert list(g.column.keys()) == ['col1','col2']
    for col, meta in cols.items():
        assert list(g.column[col].items()) == meta

def test_grid_given_column_meta_dict():
    cols = SortableDict([('col1', SortableDict([('c1m1', None), ('c1m2', None)])),
                        ('col2', SortableDict([('c2m1', None), ('c2m2', None)]))])
    g = Grid(columns=cols)
    assert list(g.column.keys()) == ['col1','col2']
    for col, meta in cols.items():
        assert list(g.column[col].items()) == list(meta.items())

def test_grid_getitem():
    g = Grid()
    g.column['test'] = {}

    row = {'test': 'This is a test'}
    g.append(row)
    assert g[0] is row

def test_grid_setitem():
    g = Grid()
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = {'test': 'This is another test'}
    g.append(row_1)
    g[0] = row_2
    assert g[0] is row_2

def test_grid_append_notdict():
    g = Grid()
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = 'This is not a dict'
    g.append(row_1)
    try:
        g.append(row_2)
        assert False
    except TypeError as e:
        assert str(e) == 'value must be a dict'
        assert len(g) == 1

def test_grid_append_v2_list_fail():
    g = Grid(version='2.0')
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = {'test': ['This should fail']}
    g.append(row_1)
    try:
        g.append(row_2)
        assert False, 'Appended invalid data type'
    except ValueError as e:
        assert str(e) == 'Data type requires version 3.0'

def test_grid_append_nover_list():
    g = Grid()
    assert g.version == Version('2.0')
    g.column['test'] = {}

    row_1 = {'test': 'This is a test'}
    row_2 = {'test': ['This should fail']}
    g.append(row_1)
    assert g.version == Version('2.0')
    g.append(row_2)
    assert g.version == Version('3.0')

def test_grid_setitem_notdict():
    g = Grid()
    g.column['test'] = {}

    row = {'test': 'This is a test'}
    g.append(row)

    try:
        g[0] = 'This is not a dict'
        assert False, 'Accepted a string'
    except TypeError:
        pass
    assert len(g) == 1
    assert g[0]['test'] == 'This is a test'

def test_grid_del():
    g = Grid()
    rows = [
            {'test': 1},
            {'test': 2},
            {'test': 3}
    ]
    g.column['test'] = {}
    g.extend(rows)
    assert len(g) == 3
    del g[1]
    assert len(g) == 2
    assert g[0] is rows[0]
    assert g[1] is rows[2]

def test_grid_insert():
    g = Grid()
    rows = [
            {'test': 1},
            {'test': 2},
            {'test': 3}
    ]
    g.column['test'] = {}
    g.extend(rows)
    assert len(g) == 3
    new_row = {'test': 'new'}
    g.insert(1, new_row)
    assert len(g) == 4
    assert g[0] is rows[0]
    assert g[1] is new_row
    assert g[2] is rows[1]
    assert g[3] is rows[2]



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Project Haystack timezone data
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si: 

import pytz
import datetime

from .version import LATEST_VER

# The official list of timezones as of 6th Jan 2016:
# Yes, that's *without* the usual country prefix.
HAYSTACK_TIMEZONES="""Abidjan
Accra
Adak
Addis_Ababa
Adelaide
Aden
Algiers
Almaty
Amman
Amsterdam
Anadyr
Anchorage
Andorra
Antananarivo
Antigua
Apia
Aqtau
Aqtobe
Araguaina
Ashgabat
Asmara
Asuncion
Athens
Atikokan
Auckland
Azores
Baghdad
Bahia
Bahia_Banderas
Bahrain
Baku
Bangkok
Barbados
Beirut
Belem
Belgrade
Belize
Berlin
Bermuda
Beulah
Bishkek
Bissau
Blanc-Sablon
Boa_Vista
Bogota
Boise
Brisbane
Broken_Hill
Brunei
Brussels
Bucharest
Budapest
Buenos_Aires
Cairo
Cambridge_Bay
Campo_Grande
Canary
Cancun
Cape_Verde
Caracas
Casablanca
Casey
Catamarca
Cayenne
Cayman
Center
Ceuta
Chagos
Chatham
Chicago
Chihuahua
Chisinau
Chita
Choibalsan
Christmas
Chuuk
Cocos
Colombo
Comoro
Copenhagen
Cordoba
Costa_Rica
Creston
Cuiaba
Curacao
Currie
Damascus
Danmarkshavn
Dar_es_Salaam
Darwin
Davis
Dawson
Dawson_Creek
Denver
Detroit
Dhaka
Dili
Djibouti
Dubai
Dublin
DumontDUrville
Dushanbe
Easter
Edmonton
Efate
Eirunepe
El_Aaiun
El_Salvador
Enderbury
Eucla
Fakaofo
Faroe
Fiji
Fortaleza
Funafuti
GMT
GMT+1
GMT+10
GMT+11
GMT+12
GMT+2
GMT+3
GMT+4
GMT+5
GMT+6
GMT+7
GMT+8
GMT+9
GMT-1
GMT-10
GMT-11
GMT-12
GMT-13
GMT-14
GMT-2
GMT-3
GMT-4
GMT-5
GMT-6
GMT-7
GMT-8
GMT-9
Galapagos
Gambier
Gaza
Gibraltar
Glace_Bay
Godthab
Goose_Bay
Grand_Turk
Guadalcanal
Guam
Guatemala
Guayaquil
Guyana
Halifax
Havana
Hebron
Helsinki
Hermosillo
Ho_Chi_Minh
Hobart
Hong_Kong
Honolulu
Hovd
Indianapolis
Inuvik
Iqaluit
Irkutsk
Istanbul
Jakarta
Jamaica
Jayapura
Jerusalem
Johannesburg
Jujuy
Juneau
Kabul
Kaliningrad
Kamchatka
Kampala
Karachi
Kathmandu
Kerguelen
Khandyga
Khartoum
Kiev
Kiritimati
Knox
Kolkata
Kosrae
Krasnoyarsk
Kuala_Lumpur
Kuching
Kuwait
Kwajalein
La_Paz
La_Rioja
Lagos
Lima
Lindeman
Lisbon
London
Lord_Howe
Los_Angeles
Louisville
Luxembourg
Macau
Maceio
Macquarie
Madeira
Madrid
Magadan
Mahe
Majuro
Makassar
Maldives
Malta
Managua
Manaus
Manila
Maputo
Marengo
Marquesas
Martinique
Matamoros
Mauritius
Mawson
Mayotte
Mazatlan
Melbourne
Mendoza
Menominee
Merida
Metlakatla
Mexico_City
Midway
Minsk
Miquelon
Mogadishu
Monaco
Moncton
Monrovia
Monterrey
Montevideo
Monticello
Montreal
Moscow
Muscat
Nairobi
Nassau
Nauru
Ndjamena
New_Salem
New_York
Nicosia
Nipigon
Niue
Nome
Norfolk
Noronha
Noumea
Novokuznetsk
Novosibirsk
Ojinaga
Omsk
Oral
Oslo
Pago_Pago
Palau
Palmer
Panama
Pangnirtung
Paramaribo
Paris
Perth
Petersburg
Phnom_Penh
Phoenix
Pitcairn
Pohnpei
Pontianak
Port-au-Prince
Port_Moresby
Port_of_Spain
Porto_Velho
Prague
Puerto_Rico
Pyongyang
Qatar
Qyzylorda
Rainy_River
Rangoon
Rankin_Inlet
Rarotonga
Recife
Regina
Rel
Resolute
Reunion
Reykjavik
Riga
Rio_Branco
Rio_Gallegos
Riyadh
Rome
Rothera
Saipan
Sakhalin
Salta
Samara
Samarkand
San_Juan
San_Luis
Santa_Isabel
Santarem
Santiago
Santo_Domingo
Sao_Paulo
Scoresbysund
Seoul
Shanghai
Simferopol
Singapore
Sitka
Sofia
South_Georgia
Srednekolymsk
St_Johns
Stanley
Stockholm
Swift_Current
Sydney
Syowa
Tahiti
Taipei
Tallinn
Tarawa
Tashkent
Tbilisi
Tegucigalpa
Tehran
Tell_City
Thimphu
Thule
Thunder_Bay
Tijuana
Tirane
Tokyo
Tongatapu
Toronto
Tripoli
Troll
Tucuman
Tunis
UCT
UTC
Ulaanbaatar
Urumqi
Ushuaia
Ust-Nera
Uzhgorod
Vancouver
Vevay
Vienna
Vientiane
Vilnius
Vincennes
Vladivostok
Volgograd
Vostok
Wake
Wallis
Warsaw
Whitehorse
Winamac
Windhoek
Winnipeg
Yakutat
Yakutsk
Yekaterinburg
Yellowknife
Yerevan
Zaporozhye
Zurich""".split('\n')
HAYSTACK_TIMEZONES_SET=set(HAYSTACK_TIMEZONES)

# Mapping of pytz-recognised timezones to Haystack timezones.
_TZ_MAP = None
_TZ_RMAP = None

def _map_timezones():
    """
    Map the official Haystack timezone list to those recognised by pytz.
    """
    tz_map = {}
    todo = HAYSTACK_TIMEZONES_SET.copy()
    for full_tz in pytz.all_timezones:
        # Finished case:
        if not bool(todo): # pragma: no cover
            # This is nearly impossible for us to cover, and an unlikely case.
            break

        # Case 1: exact match
        if full_tz in todo:
            tz_map[full_tz] = full_tz   # Exact match
            todo.discard(full_tz)
            continue

        # Case 2: suffix match after '/'
        if '/' not in full_tz:
            continue

        (prefix, suffix) = full_tz.split('/',1)
        # Case 2 exception: full timezone contains more than one '/' -> ignore
        if '/' in suffix:
            continue

        if suffix in todo:
            tz_map[suffix] = full_tz
            todo.discard(suffix)
            continue

    return tz_map

def _gen_map():
    global _TZ_MAP
    global _TZ_RMAP
    if (_TZ_MAP is None) or (_TZ_RMAP is None):
        _TZ_MAP = _map_timezones()
        _TZ_RMAP = dict([(z,n) for (n,z) in list(_TZ_MAP.items())])

def get_tz_map(version=LATEST_VER):
    """
    Return the timezone map, generating it if needed.
    """
    _gen_map()
    return _TZ_MAP

def get_tz_rmap(version=LATEST_VER):
    """
    Return the reverse timezone map, generating it if needed.
    """
    _gen_map()
    return _TZ_RMAP

def timezone(haystack_tz, version=LATEST_VER):
    """
    Retrieve the Haystack timezone
    """
    tz_map = get_tz_map(version=version)
    try:
        tz_name = tz_map[haystack_tz]
    except KeyError:
        raise ValueError('%s is not a recognised timezone on this host' \
                % haystack_tz)
    return pytz.timezone(tz_name)

def timezone_name(dt, version=LATEST_VER):
    """
    Determine an appropriate timezone for the given date/time object
    """
    tz_rmap = get_tz_rmap(version=version)
    if dt.tzinfo is None:
        raise ValueError('%r has no timezone' % dt)

    # Easy case: pytz timezone.
    try:
        tz_name = dt.tzinfo.zone
        return tz_rmap[tz_name]
    except KeyError:
        # Not in timezone map
        pass
    except AttributeError:
        # Not a pytz-compatible tzinfo
        pass

    # Hard case, try to find one that's equivalent.  Hopefully we don't get
    # many of these.  Start by getting the current timezone offset, and a
    # timezone-naïve copy of the timestamp.
    offset  = dt.utcoffset()
    dt_notz = dt.replace(tzinfo=None)

    if offset == datetime.timedelta(0):
        # UTC?
        return 'UTC'

    for olson_name, haystack_name in list(tz_rmap.items()):
        if pytz.timezone(olson_name).utcoffset(dt_notz) == offset:
            return haystack_name

    raise ValueError('Unable to get timezone of %r' % dt)



# -*- coding: utf-8 -*-
# Time zone handling/parsing tests
# (C) 2018 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

# Assume unicode literals as per Python 3
from __future__ import unicode_literals

from hszinc import zoneinfo
import iso8601
import datetime
import pytz

def test_get_tz_map():
    # This should return a mapping of all possible timezone names.
    tz_map = zoneinfo.get_tz_map()
    assert isinstance(tz_map, dict)

    # There are some timezones that aren't supported in pytz but are present
    # in Project Haystack.  Since it's impossible to know which ones, let's
    # try some well-known ones.
    assert tz_map['Adelaide'] == 'Australia/Adelaide'
    assert tz_map['Detroit'] == 'America/Detroit'
    assert tz_map['GMT'] == 'Etc/GMT'
    for etc_tz in   (('GMT',) + \
                     tuple(['GMT-%d' % n for n in range(1,13)]) + \
                     tuple(['GMT+%d' % n for n in range(1,13)])):
        assert tz_map[etc_tz] == 'Etc/%s' % etc_tz

def test_get_tz_rmap():
    # This should return a mapping of all possible timezone names.
    # The keys and values are swapped compared to get_tz_map.
    tz_rmap = zoneinfo.get_tz_rmap()
    tz_map = zoneinfo.get_tz_map()

    for iana_tz, hs_tz in tz_rmap.items():
        assert tz_map[hs_tz] == iana_tz

def test_valid_timezone():
    assert zoneinfo.timezone('Brisbane') is pytz.timezone('Australia/Brisbane')

def test_invalid_timezone():
    try:
        zoneinfo.timezone('ThisDoesNotExist')
        assert False, 'Returned a value for invalid time zone'
    except ValueError:
        pass

def test_naive_datetime():
    try:
        zoneinfo.timezone_name(datetime.datetime.now())
        assert False, 'Returned a value for naive timestamp'
    except ValueError:
        pass

def test_iso8601_datetime():
    # We don't know which one it'll pick, so pick one.
    name = zoneinfo.timezone_name(
            iso8601.parse_date('2017-01-01T00:00+10:00'))
    assert name in set([
        'Brisbane',
        'Chuuk',
        'DumontDUrville',
        'GMT-10',
        'Guam',
        'Lindeman',
        'Port_Moresby',
        'Saipan',
        'Ust-Nera',
        'Vladivostok'])

def test_oddball_iso8601_datetime():
    # We won't have a matching one for this.
    try:
        zoneinfo.timezone_name(
                iso8601.parse_date('2017-01-01T00:00+12:34'))
        assert False, 'Matched an oddball timezone'
    except ValueError:
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Zinc Grid Metadata
# (C) 2016 VRT Systems
#
# vim: set ts=4 sts=4 et tw=78 sw=4 si:

from .sortabledict import SortableDict
from .datatypes import MARKER

class MetadataObject(SortableDict):
    '''
    An object that contains some metadata fields.  Used as a convenience
    base-class for grids and columns, both of which have metadata.
    '''
    def append(self, key, value=MARKER, replace=True):
        '''
        Append the item to the metadata.
        '''
        return self.add_item(key, value, replace=replace)

    def extend(self, items, replace=True):
        '''
        Append the items to the metadata.
        '''
        if isinstance(items, dict) or isinstance(items, SortableDict):
            items = list(items.items())

        for (key, value) in items:
            self.append(key, value, replace=replace)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Pint integration helpers
# (C) 2016 VRT Systems
#

from __future__ import unicode_literals
from pint import UnitRegistry

HAYSTACK_CONVERSION = [
                    ('_', ' '),
                    ('°','deg'),
                    ('per ', '/ '),
                    ('per_h','per_hour'),
                    ('Δ','delta_'),
                    ('meters','meter'),
                    ('liters','liter'),
                    ('gallons','gallon'),
                    ('millimeters','millimeter'),
                    ('centimeters','centimeter'),
                    ('H₂O', 'H2O'),
                    ('Volt', 'volt'),
                    ('grams', 'gram'),
                    ('tons refrigeration', 'refrigeration_ton'),
                    ('%', 'percent'),
                    ('degree kelvin','degK'),
                    ('degree celsius','degC'),
                    ('degree farenheit','degF'),
                    ('pound force', 'pound_force'),
                    ('metric ton', 'metric_ton'),
                    ('fluid ounce', 'fluid_ounce'),
                    ('imperial gallon','imperial_gallon'),
                    ('galUK','UK_gallon'),
                    ('kgdegK','(kg degK)'),
                    ('tonrefh','refrigeration_ton * hour'),
                    ('tonref','refrigeration_ton'),
                    ('Nm ', 'newton meter'),
                    ('Ns', 'newton second'),
                    ('Js', 'joule second'),
                    ('short ton', 'short_ton'),
                    ('degrees angular', 'deg'),
                    ('degrees phase', 'deg'),
                    ('degPh', 'deg'),
                    ('yr','year '),
                    ('atmosphere', 'atm'),
                    ('mo','month '),
                    ('wk','week '),
                    ('parts / unit','ppu'),
                    ('parts / million','ppm'),
                    ('parts / billion','ppb'),
                    ('kcfm','kilocfm'),
                    ('kilohm','kiloohm'),
                    ('megohm','megaohm'),
                    ('volt ampere reactive', 'VAR'),
                    ('kilovolt ampere reactive', 'kVAR'),
                    ('megavolt ampere reactive', 'MVAR'),
                    ('VAh', 'volt * ampere * hour'),
                    ('kVAh', 'kilovolt * ampere * hour'),
                    ('MVAh', 'megavolt * ampere * hour'),
                    ('VARh', 'VAR * hour'),
                    ('kVARh', 'kVAR * hour'),
                    ('MVARh', 'MVAR * hour'),
                    ('hph', 'horsepower * hour'),
                    ('energy efficiency ratio', 'EER'),
                    ('coefficient of performance', 'COP'),
                    ('data center infrastructure efficiency', 'DCIE'),
                    ('power usage effectiveness', 'PUE'),
                    ('formazin nephelometric unit', 'fnu'),
                    ('nephelometric turbidity units', 'ntu'),
                    ('dBµV', 'dB microvolt'),
                    ('dBmV', 'dB millivolt'),
                    ('db','dB'),
                    ('Am', 'A * m'),
                    ('percent relative humidity', 'percentRH'),
                    ('pf', 'PF'),
                    ('power factor', 'PF'),
                    ('gH2O','g H2O'),
                    ('irradiance',''),
                    ('irr',''),
                    ('dry air', 'dry'),
                    ('dry', 'dry_air'),
                    ('kgAir','kg dry_air'),
                    ('percent obscuration', 'percentobsc'),
                    ('natural gas', ''),
                    ('Ωm', 'ohm meter'),
                    ('hecto cubic foot', 'hecto_cubic_foot'),
                    ('julian month','month'),
                    ('tenths second', 'tenths_second'),
                    ('hundredths second', 'hundredths_second'),
                    ('australian dollar','australian_dollar'),
                    ('british pound','british_pound'),
                    ('canadian dollar','canadian_dollar'),
                    ('chinese yuan','chinese_yuan'),
                    ('emerati dirham','emerati_dirham'),
                    ('indian rupee','indian_rupee'),
                    ('japanese yen','japanese_yen'),
                    ('russian ruble','russian_ruble'),
                    ('south korean won','south_korean_won'),
                    ('swedish krona','swedish_krona'),
                    ('swiss franc','swiss_franc'),
                    ('taiwan dollar','taiwan_dollar'),
                    ('us dollar','us_dollar'),
                    ('new israeli shekel','new_israeli_shekel'),
                    ('delta_K', 'delta_degC'),
                    ('delta degK', 'delta_degC'),
                    ('delta degC', 'delta_degC'),
                    ('delta degF', 'delta_degF'),
                    ('$', 'USD'),
                    ('£', 'GBP'),
                    ('元', 'CNY'),
                    ('€', 'EUR'),
                    ('₹', 'INR'),
                    ('¥', 'JPY'),
                    ('₩', 'KRW'),
                    ('of','')
                    ]
                    
PINT_CONVERSION = [
                    ('foot ** 3', 'cubic_foot'),
                    ('/','per'),
                    ('hectofoot ** 3','hecto_cubic_foot'),
                    ('meter ** 3','cubic_meter'),
                    ('Volt_per','volts_per'),
                    ('°ree','degree')
]

def to_haystack(unit):
    """
    Some parsing tweaks to fit pint units / handling of edge cases.
    """
    unit = str(unit)
    global HAYSTACK_CONVERSION
    global PINT_CONVERSION
    if unit == 'per_minute' or \
        unit == '/min' or \
        unit == 'per_second' or \
        unit == '/s' or \
        unit == 'per_hour' or \
        unit == '/h' or \
        unit == None:
        return ''
        # Those units are not units... they are impossible to fit anywhere in Pint
    
    for pint_value, haystack_value in PINT_CONVERSION:
        unit = unit.replace(pint_value, haystack_value)
    for haystack_value, pint_value in HAYSTACK_CONVERSION:
        if pint_value == '':
            continue
        unit = unit.replace(pint_value, haystack_value)
    return unit

def to_pint(unit):
    """
    Some parsing tweaks to fit pint units / handling of edge cases.
    """
    global HAYSTACK_CONVERSION
    if unit == 'per_minute' or \
        unit == '/min' or \
        unit == 'per_second' or \
        unit == '/s' or \
        unit == 'per_hour' or \
        unit == '/h' or \
        unit == None:
        return ''
        # Those units are not units... they are impossible to fit anywhere in Pint
    for haystack_value, pint_value in HAYSTACK_CONVERSION:
        unit = unit.replace(haystack_value, pint_value)
    return unit

                
def define_haystack_units():
    """
    Missing units found in project-haystack
    Added to the registry
    """
    ureg = UnitRegistry()
    ureg.define('% = [] = percent')
    ureg.define('pixel = [] = px = dot = picture_element = pel')
    ureg.define('decibel = [] = dB')
    ureg.define('ppu = [] = parts_per_unit')
    ureg.define('ppm = [] = parts_per_million')
    ureg.define('ppb = [] = parts_per_billion')
    ureg.define('%RH = [] = percent_relative_humidity = percentRH')
    ureg.define('cubic_feet = ft ** 3 = cu_ft')
    ureg.define('cfm = cu_ft * minute = liter_per_second / 0.4719475')
    ureg.define('cfh = cu_ft * hour')
    ureg.define('cfs = cu_ft * second')
    ureg.define('VAR = volt * ampere')
    ureg.define('kVAR = 1000 * volt * ampere')
    ureg.define('MVAR = 1000000 * volt * ampere')
    ureg.define('inH2O = in_H2O')
    ureg.define('dry_air = []')
    ureg.define('gas = []')
    ureg.define('energy_efficiency_ratio = [] = EER')
    ureg.define('coefficient_of_performance = [] = COP')
    ureg.define('data_center_infrastructure_efficiency = [] = DCIE')
    ureg.define('power_usage_effectiveness = [] = PUE')
    ureg.define('formazin_nephelometric_unit = [] = fnu')
    ureg.define('nephelometric_turbidity_units = [] = ntu')
    ureg.define('power_factor = [] = PF')
    ureg.define('degree_day_celsius = [] = degdaysC')
    ureg.define('degree_day_farenheit = degree_day_celsius * 9 / 5 = degdaysF')
    ureg.define('footcandle = lumen / sq_ft = ftcd')
    ureg.define('Nm = newton * meter')
    ureg.define('%obsc = [] = percent_obscuration = percentobsc')
    ureg.define('cycle = []')
    ureg.define('cph = cycle / hour')
    ureg.define('cpm = cycle / minute')
    ureg.define('cps = cycle / second')
    ureg.define('hecto_cubic_foot = 100 * cubic_foot')
    ureg.define('tenths_second = second / 10')
    ureg.define('hundredths_second = second / 100')

    #ureg.define('irradiance = W / sq_meter = irr')
    # In the definition of project haystack, there's a redundancy as irr = W/m^2
    # no need to use : watts_per_square_meter_irradiance
    
    # CURRENCY
    # I know...we won'T be able to convert right now !
    ureg.define('australian_dollar = [] = AUD')
    ureg.define('british_pound = [] = GBP = £')
    ureg.define('canadian_dollar = [] = CAD')
    ureg.define('chinese_yuan = [] = CNY = 元')
    ureg.define('emerati_dirham = [] = AED')
    ureg.define('euro = [] = EUR = €')
    ureg.define('indian_rupee = [] = INR = ₹')
    ureg.define('japanese_yen = [] = JPY = ¥')
    ureg.define('russian_ruble = [] = RUB = руб')
    ureg.define('south_korean_won = [] = KRW = ₩')
    ureg.define('swedish_krona = [] = SEK = kr')
    ureg.define('swiss_franc = [] = CHF = Fr')
    ureg.define('taiwan_dollar = [] = TWD')
    ureg.define('us_dollar = [] = USD = $')
    ureg.define('new_israeli_shekel = [] = NIS')

    return ureg
